from urllib.parse import urlparse, parse_qs, urlencode, urlunparse
import base64
import os


def xor_encode(s, key=42):
    xor_str = ''.join(chr(ord(c) ^ key) for c in s)
    encoded_bytes = base64.b64encode(xor_str.encode('utf-8'))
    return encoded_bytes.decode('utf-8')


def get_save_path(save_path: str = None) -> str:
    """
    Get the path where the file should be saved.

    Args:
        save_path (str, optional): The path to save the file.

    Returns:
        str: The path where the file is saved.
    
    Raises:
        ValueError: If save_path does not end with a '/'.
    """
    # If save_path is provided, check if it ends with a '/'
    if save_path and save_path.endswith('/'):
        return f'{save_path}'
    
    if save_path and not save_path.endswith('/'):
        raise ValueError("save_path must end with a '/'")
    
    return os.getcwd() + '/'


def get_unique_filename(save_path: str, file_name: str) -> str:
    base, ext = os.path.splitext(file_name)
    counter = 1
    new_file = f"{save_path}{base}({counter}){ext}"
    while os.path.exists(new_file):
        new_file = f"{save_path}{base}({counter}){ext}"
        counter += 1
    return new_file


def clean_data(data: dict) -> dict:
    """
    Cleans the input data by removing keys with None values.

    Args:
        data (dict): The input data.

    Returns:
        dict: The cleaned data.
    """
    return {k: v for k, v in data.items() if v is not None}

def join_url_params(base_url: str, params: dict) -> str:
    """
    Join URL with parameters while preserving existing query parameters.
    
    Args:
        base_url (str): Base URL that may contain existing parameters
        params (dict): New parameters to add
        
    Returns:
        str: URL with all parameters properly joined
    """
    # Parse the URL
    parsed = urlparse(base_url)
    
    # Get existing parameters
    existing_params = parse_qs(parsed.query)
    
    # Update with new parameters
    existing_params.update(params)
    
    # Reconstruct the URL
    return urlunparse((
        parsed.scheme,
        parsed.netloc,
        parsed.path,
        parsed.params,
        urlencode(existing_params, doseq=True),
        parsed.fragment
    ))


import time
import asyncio
from urllib.parse import urljoin
from typing import Optional, Dict, List, Optional, Union, TYPE_CHECKING
import logging

from .base import AsyncBase
from ..enums import TaskStatus

if TYPE_CHECKING:
    from . import AsyncGeoboxClient
    from .vectorlayer import AsyncVectorLayer
    from .raster import AsyncRaster
    from .model3d import AsyncModel
    from .file import AsyncFile
    from .tile3d import AsyncTile3d
    from ..api import GeoboxClient
    from ..task import Task

logger = logging.getLogger(__name__)


class AsyncTask(AsyncBase):

    BASE_ENDPOINT: str = 'tasks/'

    def __init__(self, 
        api: 'AsyncGeoboxClient', 
        uuid: str, 
        data: Optional[Dict] = {}):
        """
        Constructs all the necessary attributes for the Task object.

        Args:
            api (AsyncGeoboxClient): The API instance.
            uuid (str): The UUID of the task.
            data (Dict, optional): The task data.
        """
        super().__init__(api, uuid=uuid, data=data)
        self._data = data if isinstance(data, dict) else {}


    async def refresh_data(self) -> None:
        """
        [async] Updates the task data.
        """
        task = await self.get_task(self.api, self.uuid)
        self._data = task.data


    @property
    async def output_asset(self) -> Union['AsyncVectorLayer', 'AsyncRaster', 'AsyncModel', 'AsyncFile', 'AsyncTile3d', None]:
        """
        [async] output asset property

        Returns:
            AsyncVectorLayer | AsyncRaster | AsyncModel | AsyncFile | AsyncTile3d | None: if task type is publish, it returns the published layer

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.task import AsyncTask
            >>> async with AsyncGeoboxClient() as client:
            >>>     task = await AsyncTask.get_task(client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     await task.output_asset
        """
        if self.data.get('result', {}).get('layer_uuid'):
            return await self.api.get_vector(uuid=self.data['result']['layer_uuid'])

        elif self.data.get('result', {}).get('raster_uuid'):
            return await self.api.get_raster(uuid=self.data['result']['raster_uuid'])

        elif self.data.get('result', {}).get('model_uuid'):
            return await self.api.get_model(uuid=self.data['result']['model_uuid'])

        elif self.data.get('result', {}).get('file_uuid'):
            return await self.api.get_file(uuid=self.data['result']['file_uuid'])
        
        elif self.data.get('result', {}).get('3dtiles_uuid'):
            return await self.api.get_3dtile(uuid=self.data['result']['3dtiles_uuid'])

        else:
            return None


    @property
    def data(self) -> Dict:
        """
        Returns the task data.

        Returns:
            Dict: the task data as a dictionary
            
        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.task import AsyncTask
            >>> async with AsyncGeoboxClient() as client:
            >>>     task = await AsyncTask.get_task(client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     task.data
        """
        return self._data
    

    @data.setter
    def data(self, value: Dict) -> None:
        """
        Sets the task data.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.task import AsyncTask
            >>> async with AsyncGeoboxClient() as client:
            >>>     task = await AsyncTask.get_task(client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     task.data = {'name': 'test'}
        """
        self._data = value if isinstance(value, dict) else {}
    

    @property
    async def status(self) -> 'TaskStatus':
        """
        [async] Returns the status of the task. (auto refresh)

        Returns:
            TaskStatus: the status of the task(SUCCESS, FAILURE, ABORTED, PENDING, PROGRESS)
        
        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.task import AsyncTask
            >>> async with AsyncGeoboxClient() as client:
            >>>     task = await AsyncTask.get_task(client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     await task.status
        """
        await self.refresh_data()
        return TaskStatus(self._data.get('state'))


    @property
    def errors(self) -> Union[Dict, None]:
        """
        Get the task errors.

        Returns:
            Dict | None: if there are any errors

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.task import AsyncTask
            >>> async with AsyncGeoboxClient() as client:
            >>>     task = await AsyncTask.get_task(client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     task.errors    
        """
        result = self.data.get('result', {})
        if result.get('errors') or result.get('detail', {}).get('msg'):
            return result
        else: 
            return None


    @property
    async def progress(self) -> Union[int, None]:
        """
        [async] Returns the progress of the task.

        Returns:
            int | None: the progress of the task in percentage or None if the task is not running

        Example:
              >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.task import AsyncTask
            >>> async with AsyncGeoboxClient() as client:
            >>>     task = await AsyncTask.get_task(client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     await task.progress
        """
        endpoint = urljoin(self.endpoint, 'status/')
        response = await self.api.get(endpoint)
        
        current = response.get('current')
        total = response.get('total')
        if not total or not current:
            return None
        
        return int((current / total) * 100)

    
    async def _wait(self, timeout: Union[int, None] = None, interval: int = 1, progress_bar: bool = True) -> 'TaskStatus':
        start_time = time.time()
        last_progress = 0
        pbar = self._create_progress_bar() if progress_bar else None
        
        try:
            while True:
                self._check_timeout(start_time, timeout)
                status = await self.status
                
                if self._is_final_state(status):
                    await self._update_progress_bar(pbar, last_progress, status)
                    return status
                
                if pbar:
                    last_progress = await self._update_progress_bar(pbar, last_progress)
                
                await asyncio.sleep(interval)

        finally:
            if pbar:
                pbar.close()


    async def wait(self, timeout: Union[int, None] = None, interval: int = 1, progress_bar: bool = True, retry: int = 3) -> 'TaskStatus':
        """
        [async] Wait for the task to finish.
        
        Args:
            timeout (int, optional): Maximum time to wait in seconds.
            interval (int, optional): Time between status checks in seconds.
            progress_bar (bool, optional): Whether to show a progress bar. default: True
            retry (int, optional): Number of times to retry if waiting for the task fails. default is 3
            
        Returns:
            TaskStatus: the status of the task(SUCCESS, FAILURE, ABORTED, PENDING, PROGRESS)
            
        Raises:
            TimeoutError: If the task doesn't complete within timeout seconds.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.task import AsyncTask
            >>> async with AsyncGeoboxClient() as client:
            >>>     task = await AsyncTask.get_task(client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     await task.wait() # return the status of the task
        """
        last_exception = None

        for attempt in range(1, retry + 1):
            try:
                return await self._wait(timeout, interval, progress_bar)
            except Exception as e:
                last_exception = e
                logger.warning(f"[Retry {attempt}/{retry}] Task wait failed: {e}")
                time.sleep(interval)
        raise last_exception


    def _create_progress_bar(self) -> 'tqdm':
        """Creates a progress bar for the task."""
        try:
            from tqdm.auto import tqdm
        except ImportError:
            from .api import logger
            logger.warning("[tqdm] extra is required to show the progress bar. install with: pip insatll geobox[tqdm]")
            return None

        return tqdm(total=100, colour='green', desc=f"Task: {self.name}", unit="%", leave=True)


    def _check_timeout(self, start_time: float, timeout: Union[int, None]) -> None:
        """Checks if the task has exceeded the timeout period."""
        if timeout and time.time() - start_time > float(timeout):
            raise TimeoutError(f"Task {self.name} timed out after {timeout} seconds")


    def _is_final_state(self, status: 'TaskStatus') -> bool:
        """Checks if the task has reached a final state."""
        return status in [TaskStatus.FAILURE, TaskStatus.SUCCESS, TaskStatus.ABORTED]


    async def _update_progress_bar(self, pbar: Union['tqdm', None], last_progress: int, status: 'TaskStatus' = None) -> int:
        """
        [async] Updates the progress bar with current progress and returns the new last_progress.
        
        Args:
            pbar (tqdm | None): The progress bar to update
            last_progress (int): The last progress value
            status (TaskStatus, optional): The task status. If provided and SUCCESS, updates to 100%
            
        Returns:
            int: The new last_progress value
        """
        if not pbar:
            return last_progress
        
        if status == TaskStatus.SUCCESS:
            pbar.update(100 - last_progress)
            return 100
        
        current_progress = await self.progress
        if current_progress is not None:
            progress_diff = current_progress - last_progress
            if progress_diff > 0:
                pbar.update(progress_diff)
                return current_progress
        return last_progress


    async def abort(self) -> None:
        """
        [async] Aborts the task.

        Returns:
            None

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.task import AsyncTask
            >>> async with AsyncGeoboxClient() as client:
            >>>     task = await AsyncTask.get_task(client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     await task.abort()
        """
        endpoint = urljoin(self.endpoint, 'abort/')
        await self.api.post(endpoint)


    @classmethod
    async def get_tasks(cls, api: 'AsyncGeoboxClient', **kwargs) -> Union[List['Task'], int]:
        """
        [async] Get a list of tasks

        Args:
            api (AsyncGeoboxClient): The AsyncGeoboxClient instance for making requests.
            
        Keyword Args:
            state (TaskStatus): Available values : TaskStatus.PENDING, TaskStatus.PROGRESS, TaskStatus.SUCCESS, TaskStatus.FAILURE, TaskStatus.ABORTED
            q (str): query filter based on OGC CQL standard. e.g. "field1 LIKE '%GIS%' AND created_at > '2021-01-01'"
            search (str): search term for keyword-based searching among search_fields or all textual fields if search_fields does not have value. NOTE: if q param is defined this param will be ignored.
            search_fields (str): comma separated list of fields for searching.
            order_by (str): comma separated list of fields for sorting results [field1 A|D, field2 A|D, …]. e.g. name A, type D. NOTE: "A" denotes ascending order and "D" denotes descending order.
            return_count (bool): The count of the tasks. default is False.
            skip (int): The skip of the task. default is 0.
            limit (int): The limit of the task. default is 10.
            user_id (int): Specific user. privileges required.
            shared (bool): Whether to return shared tasks. default is False.

        Returns:
            List[AsyncTask] | int: The list of task objects or the count of the tasks if return_count is True.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.task import AsyncTask
            >>> async with AsyncGeoboxClient() as client:
            >>>     tasks = await AsyncTask.get_tasks(client)
            or  
            >>>     tasks = await client.get_tasks()
        """
        params = {
            'f': 'json',
            'state': kwargs.get('state').value if kwargs.get('state') else None,
            'q': kwargs.get('q'),
            'search': kwargs.get('search'),
            'search_fields': kwargs.get('search_fields'),
            'order_by': kwargs.get('order_by'),
            'return_count': kwargs.get('return_count', False),
            'skip': kwargs.get('skip'),
            'limit': kwargs.get('limit'),
            'user_id': kwargs.get('user_id'),
            'shared': kwargs.get('shared', False)
        }
        return await super()._get_list(api, cls.BASE_ENDPOINT, params, factory_func=lambda api, item: AsyncTask(api, item['uuid'], item))
        


    @classmethod
    async def get_task(cls, api: 'AsyncGeoboxClient', uuid: str) -> 'AsyncTask':
        """
        [async] Gets a task.

        Args:
            api (AsyncGeoboxClient): The AsyncGeoboxClient instance for making requests.
            uuid (str): The UUID of the task.

        Returns:
            AsyncTask: The task object.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.task import AsyncTask
            >>> async with AsyncGeoboxClient() as client:
            >>>     task = await AsyncTask.get_task(client, uuid="12345678-1234-5678-1234-567812345678")
            or  
            >>>     task = await client.get_task(uuid="12345678-1234-5678-1234-567812345678")
        """
        return await super()._get_detail(api, cls.BASE_ENDPOINT, uuid, factory_func=lambda api, item: AsyncTask(api, item['uuid'], item))


    def to_sync(self, sync_client: 'GeoboxClient') -> 'Task':
        """
        Switch to sync version of the task instance to have access to the sync methods

        Args:
            sync_client (SyncGeoboxClient): The sync version of the GeoboxClient instance for making requests.

        Returns:
            Task: the sync instance of the task.

        Example:
            >>> from geobox import Geoboxclient
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.task import AsyncTask
            >>> client = GeoboxClient()
            >>> async with AsyncGeoboxClient() as async_client:
            >>>     task = await AsyncTask.get_task(async_client, uuid="12345678-1234-5678-1234-567812345678")
            or  
            >>>     task = await async_client.get_task(uuid="12345678-1234-5678-1234-567812345678")
            >>>     sync_task = task.to_sync(client)
        """
        from ..task import Task

        return Task(api=sync_client, uuid=self.uuid, data=self.data)


import pytest
from urllib.parse import urljoin
from unittest.mock import patch

from geobox.tileset import Tileset
from geobox.enums import LayerType
from geobox.user import User
from geobox.task import Task
from geobox.vectorlayer import VectorLayer
from geobox.view import VectorLayerView


def test_init(api, mock_tileset_data):
    """Test Tileset initialization."""
    tileset = Tileset(api, uuid=mock_tileset_data['uuid'], data=mock_tileset_data)
    assert tileset.uuid == mock_tileset_data['uuid']
    assert tileset.data == mock_tileset_data
    assert tileset.endpoint == f'{Tileset.BASE_ENDPOINT}{mock_tileset_data["uuid"]}/'


def test_repr(api, mock_tileset_data):
    """Test Tileset string representation."""
    tileset = Tileset(api, uuid=mock_tileset_data['uuid'], data=mock_tileset_data)
    assert repr(tileset) == f"Tileset(uuid={tileset.uuid}, name={mock_tileset_data['name']})"


def test_getattr(api, mock_tileset_data):
    """Test Tileset attribute access."""
    tileset = Tileset(api, uuid=mock_tileset_data['uuid'], data=mock_tileset_data)
    assert tileset.name == mock_tileset_data['name']
    assert tileset.display_name == mock_tileset_data['display_name']
    assert tileset.description == mock_tileset_data['description']
    assert tileset.min_zoom == mock_tileset_data['min_zoom']
    assert tileset.max_zoom == mock_tileset_data['max_zoom']
    assert tileset.state == mock_tileset_data['state']
    with pytest.raises(AttributeError):
        _ = tileset.nonexistent_attribute


def test_create_tileset(api, mock_tileset_data, mock_vector_data, mock_view_data):
    """Test creating a tileset"""
    api.post.return_value = mock_tileset_data
    layers = [VectorLayer(api, uuid=mock_vector_data['uuid'], layer_type=LayerType.MultiPoint, data=mock_vector_data), 
                VectorLayerView(api, uuid=mock_view_data['uuid'], layer_type=LayerType.MultiPoint, data=mock_view_data)]
    tileset = Tileset.create_tileset(api, 
                                        mock_tileset_data['name'], 
                                        layers=layers,
                                        display_name=mock_tileset_data['display_name'], 
                                        description=mock_tileset_data['description'], 
                                        min_zoom=mock_tileset_data['min_zoom'], 
                                        max_zoom=mock_tileset_data['max_zoom'])
    assert isinstance(tileset, Tileset)
    assert tileset.uuid == mock_tileset_data['uuid']
    assert tileset.data == mock_tileset_data
    api.post.assert_called_once_with('tilesets/', {'name': mock_tileset_data['name'], 
                                                    'min_zoom': mock_tileset_data['min_zoom'], 
                                                    'max_zoom': mock_tileset_data['max_zoom'], 
                                                    'layers': [{'layer_type': 'vector', 
                                                                'layer_uuid': mock_vector_data['uuid']}, 
                                                                {'layer_type': 'view', 
                                                                'layer_uuid': mock_view_data['uuid']}]})


def test_get_tilesets(api, mock_tileset_data):
    """Test getting tilesets"""
    tilesets = [mock_tileset_data for _ in range(3)]
    with patch.object(api, 'get', return_value=tilesets):
        tilesets_list = Tileset.get_tilesets(api)
        assert tilesets_list is not None
        assert len(tilesets_list) == 3
        assert isinstance(tilesets_list[0], Tileset)
        assert tilesets_list[0].uuid == mock_tileset_data['uuid']
        assert tilesets_list[0].data == mock_tileset_data


def test_get_tilesets_by_ids(api, mock_tileset_data):
    api.get.return_value = [{**mock_tileset_data, **{'id': 1}}, {**mock_tileset_data, **{'id': 2}}]
    tilesets = Tileset.get_tilesets_by_ids(api, [1, 2])
    assert isinstance(tilesets[0], Tileset)
    assert tilesets[0].uuid == mock_tileset_data['uuid']
    assert tilesets[0].data == {**mock_tileset_data, **{'id': 1}}


def test_get_tileset(api, mock_tileset_data):
    """Test getting tileset by UUID."""
    with patch.object(api, 'get', return_value=mock_tileset_data):
        tileset = Tileset.get_tileset(api, uuid=mock_tileset_data['uuid'])
        assert isinstance(tileset, Tileset)
        assert tileset.uuid == mock_tileset_data['uuid']
        assert tileset.name == mock_tileset_data['name']
        api.get.assert_called_once_with(tileset.endpoint)


def test_get_tileset_by_name(api, mock_tileset_data):
    """Test getting tileset by UUID."""
    with patch.object(api, 'get', return_value=[mock_tileset_data]):
        tileset = Tileset.get_tileset_by_name(api, name=mock_tileset_data['name'])
        assert isinstance(tileset, Tileset)
        assert tileset.uuid == mock_tileset_data['uuid']
        assert tileset.name == mock_tileset_data['name']
        api.get.assert_called_once_with(f"{Tileset.BASE_ENDPOINT}?f=json&q=name+%3D+%27test_tile%27&return_count=False&skip=0&limit=10&shared=False")
        # not found
        tileset = Tileset.get_tileset_by_name(api, name='not_found')
        assert tileset == None


def test_update(api, mock_tileset_data):
    """Test updating tileset properties."""
    updated_data = {
        'name': 'updated_name',
        'display_name': 'Updated Tileset',
        'description': 'Updated description',
        'min_zoom': 1,
        'max_zoom': 15
    }
    with patch.object(api, 'put', return_value=updated_data):
        tileset = Tileset(api, uuid=mock_tileset_data['uuid'], data=mock_tileset_data)
        tileset.update(
            name='updated_name',
            display_name='Updated Tileset',
            description='Updated description',
            min_zoom=1,
            max_zoom=15
        )
        assert tileset.name == 'updated_name'
        assert tileset.display_name == 'Updated Tileset'
        assert tileset.description == 'Updated description'
        assert tileset.min_zoom == 1
        assert tileset.max_zoom == 15


def test_delete(api, mock_tileset_data):
    """Test deleting tileset."""
    with patch.object(api, 'delete'):
        tileset = Tileset(api, uuid=mock_tileset_data['uuid'], data=mock_tileset_data)
        endpoint = tileset.endpoint
        tileset.delete()
        api.delete.assert_called_once_with(endpoint)


def test_get_layers(api, mock_tileset_data, mock_vector_data, mock_view_data):
    """Test getting tileset layers."""
    mock_layers = [
        mock_vector_data,
        mock_view_data
    ]
    with patch.object(api, 'get', return_value=mock_layers):
        tileset = Tileset(api, uuid=mock_tileset_data['uuid'], data=mock_tileset_data)
        layers = tileset.get_layers()
        assert len(layers) == 2
        assert layers[0].name == mock_vector_data['name']
        assert type(layers[0]) == VectorLayer
        assert layers[1].name == mock_view_data['name']
        assert type(layers[1]) == VectorLayerView

    # return count
    with patch.object(api, 'get', return_value=10):
        tileset = Tileset(api, uuid=mock_tileset_data['uuid'], data=mock_tileset_data)
        layer_count = tileset.get_layers(return_count=True)
        assert layer_count == 10
        api.get.assert_called_once_with(f'{tileset.endpoint}layers/?f=json&return_count=True&skip=0&limit=10&shared=False')


def test_add_layer(api, mock_tileset_data, mock_vector_data, mock_view_data):
    """Test adding layer to tileset."""
    with patch.object(api, 'post'):
        tileset = Tileset(api, uuid=mock_tileset_data['uuid'], data=mock_tileset_data)
        layer = VectorLayer(api, uuid=mock_vector_data['uuid'], layer_type=LayerType.MultiPoint, data=mock_vector_data)
        view = VectorLayerView(api, uuid=mock_view_data['uuid'], layer_type=LayerType.MultiPoint, data=mock_view_data)
        tileset.add_layer(layer=layer)
        endpoint = urljoin(tileset.endpoint, 'layers/')
        api.post.assert_called_once_with(endpoint, {'layer_type': 'vector', 'layer_uuid': mock_vector_data['uuid']}, is_json=False)
        api.post.reset_mock()
        tileset.add_layer(layer=view)
        api.post.assert_called_once_with(endpoint, {'layer_type': 'view', 'layer_uuid': mock_view_data['uuid']}, is_json=False)

        # usupported input
        with pytest.raises(ValueError):
            tileset.add_layer(layer='')


def test_delete_layer(api, mock_tileset_data, mock_vector_data, mock_view_data):
    """Test deleting layer from tileset."""
    with patch.object(api, 'delete'):
        tileset = Tileset(api, uuid=mock_tileset_data['uuid'], data=mock_tileset_data)
        layer = VectorLayer(api, uuid=mock_vector_data['uuid'], layer_type=LayerType.MultiPoint, data=mock_vector_data)
        view = VectorLayerView(api, uuid=mock_view_data['uuid'], layer_type=LayerType.MultiPoint, data=mock_view_data)
        tileset.delete_layer(layer=layer)
        endpoint = urljoin(tileset.endpoint, 'layers/')
        api.delete.assert_called_once_with(endpoint, {'layer_type': 'vector', 'layer_uuid': mock_vector_data['uuid']}, is_json=False)
        api.delete.reset_mock()
        tileset.delete_layer(layer=view)
        api.delete.assert_called_once_with(endpoint, {'layer_type': 'view', 'layer_uuid': mock_view_data['uuid']}, is_json=False)

        # usupported input
        with pytest.raises(ValueError):
            tileset.delete_layer(layer='')


def test_share(api, mock_tileset_data, mock_user_data):
    """Test file sharing."""
    with patch.object(api, 'post'):
        tileset = Tileset(api, uuid=mock_tileset_data['uuid'], data=mock_tileset_data)
        users = [
            User(api, user_id=mock_user_data['1']['id'], data=mock_user_data['1']),
            User(api, user_id=mock_user_data['2']['id'], data=mock_user_data['2'])
        ]
        tileset.share(users=users)
        api.post.assert_called_once_with(
            f'{tileset.endpoint}share/',
            {'user_ids': [1, 2]}, is_json=False
        )

def test_unshare(api, mock_tileset_data, mock_user_data):
    """Test file unsharing."""
    with patch.object(api, 'post'):
        tileset = Tileset(api, uuid=mock_tileset_data['uuid'], data=mock_tileset_data)
        users = [
            User(api, user_id=mock_user_data['1']['id'], data=mock_user_data['1']),
            User(api, user_id=mock_user_data['2']['id'], data=mock_user_data['2'])
        ]
        tileset.unshare(users=users)
        api.post.assert_called_once_with(
            f'{tileset.endpoint}unshare/',
            {'user_ids': [1, 2]}, is_json=False
        )


def test_get_shared_users(api, mock_tileset_data, mock_user_data):
    """Test getting shared users."""
    mock_response = [
        mock_user_data['1'],
        mock_user_data['2']
    ]
    
    with patch.object(api, 'get', return_value=mock_response):
        tileset = Tileset(api, uuid=mock_tileset_data['uuid'], data=mock_tileset_data)
        result = tileset.get_shared_users(search='user', limit=2)
        
        api.get.assert_called_once_with(
            f'{tileset.endpoint}shared-with-users/?search=user&skip=0&limit=2'
        )

        assert len(result) == 2
        assert result[0].first_name == 'test 1'
        assert result[0].last_name == 'test 1'
        assert result[1].first_name == 'test 2'
        assert result[1].last_name == 'test 2'


def test_get_tile_json(api, mock_tileset_data):
    """Test getting tile JSON."""
    mock_tilejson = {
        'tiles': ['https://api.geobox.ir/v1/tilesets/test-uuid-123/tiles/{z}/{x}/{y}.pbf'],
        'minzoom': 0,
        'maxzoom': 14
    }
    with patch.object(api, 'get', return_value=mock_tilejson):
        tileset = Tileset(api, uuid=mock_tileset_data['uuid'], data=mock_tileset_data)
        tilejson = tileset.get_tile_json()
        assert tilejson == mock_tilejson
        endpoint = urljoin(tileset.endpoint, 'tilejson.json/')
        api.get.assert_called_once_with(endpoint)


def test_update_tileset_extent(api, mock_tileset_data):
    """Test updating tileset extent."""
    with patch.object(api, 'post', return_value={'message': 'Extent updated'}):
        tileset = Tileset(api, uuid=mock_tileset_data['uuid'], data=mock_tileset_data)
        result = tileset.update_tileset_extent()
        assert result == {'message': 'Extent updated'}
        endpoint = urljoin(tileset.endpoint, 'updateExtent/')
        api.post.assert_called_once_with(endpoint)


def test_get_tile_pbf_url(api, mock_tileset_data):
    """Test getting tile from tileset."""
    tileset = Tileset(api, uuid=mock_tileset_data['uuid'], data=mock_tileset_data)
    url = tileset.get_tile_pbf_url(x=1, y=1, z=1)
    assert url == f"{api.base_url}{tileset.endpoint}tiles/1/1/1.pbf"
    # apikey
    api.access_token = ''
    api.apikey = 'apikey_1234'
    url = tileset.get_tile_pbf_url(x=1, y=1, z=1)
    assert url == f"{api.base_url}{tileset.endpoint}tiles/1/1/1.pbf?apikey=apikey_1234"


def test_seed_cache(api, mock_tileset_data, mock_success_task_data):
    tileset = Tileset(api, uuid=mock_tileset_data['uuid'], data=mock_tileset_data)
    api.post.return_value = [{'task_id': mock_success_task_data['id']}, {'task_id': mock_success_task_data['id']}]
    api.get_task.return_value = Task(api, mock_success_task_data['uuid'], mock_success_task_data)

    expected_data = {
        "from_zoom": 0,
        "workers": 4,
        "to_zoom": 10,
        "extent": [10, 20, 30, 40],
        "user_id": 0
    }

    api.get.return_value = mock_success_task_data
    task = tileset.seed_cache(from_zoom=0, workers=4, to_zoom=10, extent=[10, 20, 30, 40])[0]
    assert task.id == mock_success_task_data['id']
    api.post.assert_called_once_with(f'{tileset.endpoint}cache/seed/', expected_data)


def test_update_cache(api, mock_tileset_data, mock_success_task_data):
    tileset = Tileset(api, uuid=mock_tileset_data['uuid'], data=mock_tileset_data)
    task_data = {'task_id': mock_success_task_data['id'], 'status': 'pending'}
    api.post.return_value = [{'task_id': mock_success_task_data['id']}]
    api.get_task.return_value = Task(api, mock_success_task_data['uuid'], mock_success_task_data)

    api.get.return_value = mock_success_task_data

    task = tileset.update_cache(from_zoom=0, to_zoom=2, extents=[[1, 2, 3, 4]], user_id=1)[0]
    assert isinstance(task, Task)
    api.post.assert_called_once_with(f'{tileset.endpoint}cache/update/', {'from_zoom': 0, 'to_zoom': 2, 'extents': [[1, 2, 3, 4]], 'user_id': 1})
    # error
    api.post.return_value = [1, 2, 3]

    api.get.return_value = mock_success_task_data

    with pytest.raises(ValueError, match="Failed to update cache"):
        task = tileset.update_cache(from_zoom=0, to_zoom=2)
        

def test_clear_cache(api, mock_tileset_data):
    """Test clearing tileset cache."""
    with patch.object(api, 'post'):
        tileset = Tileset(api, uuid=mock_tileset_data['uuid'], data=mock_tileset_data)
        result = tileset.clear_cache()
        endpoint = urljoin(tileset.endpoint, 'cache/clear/')
        api.post.assert_called_once_with(endpoint) 


def test_cache_size(api, mock_tileset_data):
    tileset = Tileset(api, uuid=mock_tileset_data['uuid'], data=mock_tileset_data)
    api.post.return_value = 1024
    size = tileset.cache_size
    assert size == 1024
    api.post.assert_called_once_with(f'{tileset.endpoint}cache/size/')


def test_to_async(api, async_api, mock_tileset_data):
    tileset = Tileset(api, uuid=mock_tileset_data['uuid'], data=mock_tileset_data)
    async_instance = tileset.to_async(async_api)
    assert async_instance.api == async_api


import pytest
from unittest.mock import patch

from geobox.raster_analysis import RasterAnalysis
from geobox.enums import FieldType, LayerType
from geobox.field import Field
from geobox.raster import Raster
from geobox.task import Task
from geobox.vectorlayer import VectorLayer


def test_init(api):
    analysis = RasterAnalysis(api)
    assert type(analysis) == RasterAnalysis
    assert analysis.api == api


def test_repr(api):
    analysis = RasterAnalysis(api)
    assert repr(analysis) == 'RasterAnalysis()'


def test_rasterize(api, mock_success_task_data, mock_vector_data):
    task = Task(api, uuid=mock_success_task_data['uuid'], data=mock_success_task_data)
    api.get_task.return_value = task
    api.post.return_value = {
        'task_id': mock_success_task_data['uuid']
    }

    vector = VectorLayer(api, uuid=mock_vector_data['uuid'], layer_type=LayerType.Polygon, data=mock_vector_data)

    raster_analysis = RasterAnalysis(api)
    task = raster_analysis.rasterize(
        layer=vector,
        output_raster_name='test'
    )

    assert type(task) == Task
    assert task.uuid == mock_success_task_data['uuid']
    api.post.assert_called_once_with(endpoint='analysis/rasterize/', payload={'layer_uuid': vector.uuid, 'output_raster_name': 'test', 'is_view': False, 'pixel_size': 10, 'nodata': -9999, 'data_type': 'int16', 'burn_value': 1}, is_json=False)

    # error
    with pytest.raises(ValueError):
        raster_analysis.rasterize(
            layer='vector',
            output_raster_name='test'
        )


def test_polygonize(api, mock_success_task_data, mock_raster_data):
    task = Task(api, uuid=mock_success_task_data['uuid'], data=mock_success_task_data)
    api.get_task.return_value = task
    api.post.return_value = {
        'task_id': mock_success_task_data['uuid']
    }

    raster = Raster(api, uuid=mock_raster_data['uuid'], data=mock_raster_data)

    raster_analysis = RasterAnalysis(api)
    task = raster_analysis.polygonize(
        raster=raster,
        output_layer_name='test'
    )

    assert type(task) == Task
    assert task.uuid == mock_success_task_data['uuid']
    api.post.assert_called_once_with(endpoint='analysis/polygonize/', payload={'raster_uuid': raster.uuid, 'output_layer_name': 'test', 'band_index': 1, 'mask_nodata': False, 'connectivity': 4}, is_json=False)


def test_clip(api, mock_success_task_data, mock_raster_data, mock_vector_data):
    task = Task(api, uuid=mock_success_task_data['uuid'], data=mock_success_task_data)
    api.get_task.return_value = task
    api.post.return_value = {
        'task_id': mock_success_task_data['uuid']
    }

    vector = VectorLayer(api, uuid=mock_vector_data['uuid'], layer_type=LayerType.Polygon, data=mock_vector_data)
    raster = Raster(api, uuid=mock_raster_data['uuid'], data=mock_raster_data)

    raster_analysis = RasterAnalysis(api)
    task = raster_analysis.clip(
        layer=vector,
        raster=raster,
        output_raster_name='test'
    )

    assert type(task) == Task
    assert task.uuid == mock_success_task_data['uuid']
    api.post.assert_called_once_with(endpoint='analysis/clip/', payload={'raster_uuid': raster.uuid, 'layer_uuid': '297fa7ca-877a-400c-8003-d65de9e791c2', 'output_raster_name': 'test', 'is_view': False, 'dst_nodata': -9999, 'crop': True, 'resample': 'near'}, is_json=False)

    # error
    with pytest.raises(ValueError):
        raster_analysis.clip(
            layer='vector',
            raster=raster,
            output_raster_name='test'
        ) 


def test_calculator(api, mock_success_task_data, mock_raster_data):
    task = Task(api, uuid=mock_success_task_data['uuid'], data=mock_success_task_data)
    api.get_task.return_value = task
    api.post.return_value = {
        'task_id': mock_success_task_data['uuid']
    }

    raster = Raster(api, uuid=mock_raster_data['uuid'], data=mock_raster_data)

    raster_analysis = RasterAnalysis(api)
    task = raster_analysis.calculator(
        variables={"RED": raster.uuid},
        expr='where(SLOPE>30,1,0)',
        output_raster_name='test'
    )

    assert type(task) == Task
    assert task.uuid == mock_success_task_data['uuid']
    api.post.assert_called_once_with(endpoint='analysis/calculator/', payload={'variables': {'RED': raster.uuid}, 'expr': 'where(SLOPE>30,1,0)', 'output_raster_name': 'test', 'resample': 'bilinear', 'out_dtype': 'float32', 'dst_nodata': -9999}, is_json=False)


def test_slope(api, mock_success_task_data, mock_raster_data):
    task = Task(api, uuid=mock_success_task_data['uuid'], data=mock_success_task_data)
    api.get_task.return_value = task
    api.post.return_value = {
        'task_id': mock_success_task_data['uuid']
    }

    raster = Raster(api, uuid=mock_raster_data['uuid'], data=mock_raster_data)

    raster_analysis = RasterAnalysis(api)
    task = raster_analysis.slope(
        raster=raster,
        output_raster_name='test'
    )

    assert type(task) == Task
    assert task.uuid == mock_success_task_data['uuid']
    api.post.assert_called_once_with(endpoint='analysis/slope/', payload={'raster_uuid': raster.uuid, 'output_raster_name': 'test', 'slope_units': 'degree', 'algorithm': 'Horn', 'scale': 1, 'compute_edges': True, 'nodata_out': -9999}, is_json=False)


def test_aspect(api, mock_success_task_data, mock_raster_data):
    task = Task(api, uuid=mock_success_task_data['uuid'], data=mock_success_task_data)
    api.get_task.return_value = task
    api.post.return_value = {
        'task_id': mock_success_task_data['uuid']
    }

    raster = Raster(api, uuid=mock_raster_data['uuid'], data=mock_raster_data)

    raster_analysis = RasterAnalysis(api)
    task = raster_analysis.aspect(
        raster=raster,
        output_raster_name='test'
    )

    assert type(task) == Task
    assert task.uuid == mock_success_task_data['uuid']
    api.post.assert_called_once_with(endpoint='analysis/aspect/', payload={'raster_uuid': raster.uuid, 'output_raster_name': 'test', 'algorithm': 'Horn', 'trigonometric': False, 'zero_for_flat': True, 'compute_edges': True, 'nodata_out': -9999}, is_json=False)


def test_reclassify(api, mock_success_task_data, mock_raster_data):
    task = Task(api, uuid=mock_success_task_data['uuid'], data=mock_success_task_data)
    api.get_task.return_value = task
    api.post.return_value = {
        'task_id': mock_success_task_data['uuid']
    }

    raster = Raster(api, uuid=mock_raster_data['uuid'], data=mock_raster_data)

    raster_analysis = RasterAnalysis(api)
    task = raster_analysis.reclassify(
        raster=raster,
        output_raster_name='test',
        rules='{"1": 10, "2": 20, "3": 30}'
    )

    assert type(task) == Task
    assert task.uuid == mock_success_task_data['uuid']
    api.post.assert_called_once_with(endpoint='analysis/reclassify/', payload={'raster_uuid': raster.uuid, 'output_raster_name': 'test', 'rules': '{"1": 10, "2": 20, "3": 30}', 'nodata_in': -9999, 'nodata_out': -9999, 'out_dtype': 'int16', 'inclusive': 'left'}, is_json=False)


def test_resample(api, mock_success_task_data, mock_raster_data):
    task = Task(api, uuid=mock_success_task_data['uuid'], data=mock_success_task_data)
    api.get_task.return_value = task
    api.post.return_value = {
        'task_id': mock_success_task_data['uuid']
    }

    raster = Raster(api, uuid=mock_raster_data['uuid'], data=mock_raster_data)

    raster_analysis = RasterAnalysis(api)
    task = raster_analysis.resample(
        raster=raster,
        output_raster_name='test',
        out_res='10,10'
    )

    assert type(task) == Task
    assert task.uuid == mock_success_task_data['uuid']
    api.post.assert_called_once_with(endpoint='analysis/resample/', payload={'raster_uuid': raster.uuid, 'output_raster_name': 'test', 'out_res': '10,10', 'resample_method': 'near', 'dst_nodata': -9999}, is_json=False)

    # error
    with pytest.raises(ValueError):
        raster_analysis.resample(
            raster=raster,
            output_raster_name='test'
        )


def test_idw_interpolation(api, mock_success_task_data, mock_vector_data, mock_field_data):
    task = Task(api, uuid=mock_success_task_data['uuid'], data=mock_success_task_data)
    api.get_task.return_value = task
    api.post.return_value = {
        'task_id': mock_success_task_data['uuid']
    }

    vector = VectorLayer(api, uuid=mock_vector_data['uuid'], layer_type=LayerType.Polygon, data=mock_vector_data)
    field = Field(layer=vector, data_type=FieldType.String, field_id=mock_field_data['id'], data=mock_field_data)

    raster_analysis = RasterAnalysis(api)
    task = raster_analysis.idw_interpolation(
        layer=vector,
        output_raster_name='test',
        z_field=field
    )

    assert type(task) == Task
    assert task.uuid == mock_success_task_data['uuid']
    api.post.assert_called_once_with(endpoint='analysis/idw/', payload={'layer_uuid': vector.uuid, 'output_raster_name': 'test', 'z_field': 'test', 'is_view': False, 'pixel_size': 10, 
        'power': 2.0, 'smoothing': 0.0, 'max_points': 16, 'radius': 1000, 'nodata': -9999, 'out_dtype': 'float32'}, is_json=False)


def test_constant(api, mock_success_task_data):
    task = Task(api, uuid=mock_success_task_data['uuid'], data=mock_success_task_data)
    api.get_task.return_value = task
    api.post.return_value = {
        'task_id': mock_success_task_data['uuid']
    }


    raster_analysis = RasterAnalysis(api)
    task = raster_analysis.constant(
        output_raster_name='test',
        extent='0,0,100,100', 
        value=10
    )

    assert type(task) == Task
    assert task.uuid == mock_success_task_data['uuid']
    api.post.assert_called_once_with(endpoint='analysis/constant/', payload={'output_raster_name': 'test', 'extent': '0,0,100,100', 'value': 10, 'pixel_size': 10, 'dtype': 'float32', 'nodata': -9999}, is_json=False)


def test_fill_nodata(api, mock_success_task_data, mock_raster_data):
    task = Task(api, uuid=mock_success_task_data['uuid'], data=mock_success_task_data)
    api.get_task.return_value = task
    api.post.return_value = {
        'task_id': mock_success_task_data['uuid']
    }

    raster = Raster(api, uuid=mock_raster_data['uuid'], data=mock_raster_data)

    raster_analysis = RasterAnalysis(api)
    task = raster_analysis.fill_nodata(
        raster=raster,
        output_raster_name='test'
    )

    assert type(task) == Task
    assert task.uuid == mock_success_task_data['uuid']
    api.post.assert_called_once_with(endpoint='analysis/fill/', payload={'raster_uuid': raster.uuid, 'output_raster_name': 'test', 'band': 1}, is_json=False)


def test_proximity(api, mock_success_task_data, mock_raster_data):
    task = Task(api, uuid=mock_success_task_data['uuid'], data=mock_success_task_data)
    api.get_task.return_value = task
    api.post.return_value = {
        'task_id': mock_success_task_data['uuid']
    }

    raster = Raster(api, uuid=mock_raster_data['uuid'], data=mock_raster_data)

    raster_analysis = RasterAnalysis(api)
    task = raster_analysis.proximity(
        raster=raster,
        output_raster_name='test'
    )

    assert type(task) == Task
    assert task.uuid == mock_success_task_data['uuid']
    api.post.assert_called_once_with(endpoint='analysis/proximity/', payload={'raster_uuid': 'aea9c935-b0e2-423d-b25d-6c41d5ac22c5', 'output_raster_name': 'test', 'dist_units': 'GEO', 'burn_value': 1, 'nodata': -9999}, is_json=False)


import pytest
from unittest.mock import patch

from geobox.log import Log
from geobox.user import User

def test_init(api, mock_log_data):
    log = Log(api, mock_log_data['id'], mock_log_data)
    assert log.log_id == mock_log_data['id']
    assert log.data == mock_log_data
    assert log.endpoint == f'{Log.BASE_ENDPOINT}{log.log_id}'


def test_repr(api, mock_log_data):
    log = Log(api, mock_log_data['id'], mock_log_data)
    assert repr(log) == f"Log(id={log.log_id}, activity_type={log.activity_type})"


def test_get_logs(api, mock_log_data):
    api.get.return_value = [mock_log_data, mock_log_data]
    logs = Log.get_logs(api)
    api.get.assert_called_once_with(f'{Log.BASE_ENDPOINT}')
    assert len(logs) == 2
    assert type(logs[0]) == Log
    assert logs[0].data == mock_log_data


def test_delete(api, mock_log_data):
    log = Log(api, mock_log_data['id'], mock_log_data)
    endpoint = log.endpoint
    log.delete()
    api.delete.assert_called_once_with(endpoint)
    assert log.log_id is None
    assert log.endpoint is None


def test_user_property(api, mock_log_data, mock_user_data):
    log = Log(api, mock_log_data['id'], mock_log_data)
    api.get_user.return_value = User(api, mock_user_data['1']['id'], mock_user_data['1'])
    user = log.user
    assert user.data == mock_user_data['1']


def test_to_async(api, async_api, mock_log_data):
    log = Log(api, mock_log_data['id'], mock_log_data)
    async_instance = log.to_async(async_api)
    assert async_instance.api == async_api 


from enum import Enum

# File Enums
class InputGeomType(Enum):
    POINT = 'POINT'
    LINESTRING = 'LINESTRING'
    POLYGON = 'POLYGON'
    MULTIPOINT = 'MULTIPOINT'
    MULTILINESTRING = 'MULTILINESTRING'
    MULTIPOLYGON = 'MULTIPOLYGON'
    POINT_Z = 'POINTZ'
    LINESTRING_Z = 'LINESTRINGZ'
    POLYGON_Z = 'POLYGONZ'
    MULTIPOINT_Z = 'MULTIPOINTZ'
    MULTILINESTRING_Z = 'MULTILINESTRINGZ'
    MULTIPOLYGON_Z = 'MULTIPOLYGONZ'
    
class PublishFileType(Enum):
    VECTOR = 'vector'
    RASTER = 'raster'
    MODEL3D = 'model3d'
    Tiles3D = 'tiles3d'

class FileType(Enum):
    Compressed = 'Compressed'
    Complex = 'Complex'
    Image = 'Image'
    Video = 'Video'
    Document = 'Document'
    GPKG = 'GPKG'
    DXF = 'DXF'
    GPX = 'GPX'
    CSV = 'CSV'
    Shapefile = 'Shapefile'
    KML = 'KML'
    GLB = 'GLB'
    FileGDB = 'FileGDB'
    GeoTIFF = 'GeoTIFF'
    GeoJSON = 'GeoJSON'
    ThreedTiles = 'ThreedTiles'

class FileFormat(Enum):
    # Spatial Data Formats
    Shapefile = '.shp'
    FileGDB = '.gdb'
    CSV = '.csv'
    KML = '.kml'
    DXF = '.dxf'
    GPX = '.gpx'
    GPKG = '.gpkg'
    GeoJSON = '.geojson'
    # GeoTIFF = '.tiff'

    # Image Formats
    JPG = '.jpg'
    JPEG = '.jpeg'
    SVG = '.svg'
    PNG = '.png'
    BMP = '.bmp'
    TIF = '.tif'
    GIF = '.gif'

    # Video Formats
    MP4 = '.mp4'
    AVI = '.avi'
    OGG = '.ogg'
    WEBM = '.webm'
    MPEG = '.mpeg'
    WMV = '.wmv'

    # Document Formats
    TXT = '.txt'
    DOC = '.doc'
    DOCX = '.docx'
    XLS = '.xls'
    XLSX = '.xlsx'
    PPT = '.ppt'
    PPTX = '.pptx'
    PDF = '.pdf'

    # Compressed Formats
    ZIP = '.zip'
    SEVEN_Z = '.7z'  # 7z
    RAR = '.rar'
    ISO = '.iso'
    GZ = '.gz'
    BZ2 = '.bz2'
    TAR = '.tar'
    CAB = '.cab'
    IMG = '.img'


# Field Enums
class FieldType(Enum):
    """
    A class representing the data type of a field.
    """
    Integer = "Integer"
    """
    An integer field.
    """
    Float = "Float"
    """
    A float field.
    """
    String = "String"
    """
    A string field.
    """
    Long = "Long"
    """
    A Long field.
    """
    DATE = "Date"
    """
    A Date field.
    """
    TIME = "Time"
    """
    A Time Field.
    """
    DATETIME = "DateTime"
    """
    A DateTime Field.
    """


# Layer Enums
class LayerType(Enum):
    """
    Enumeration of supported vector layer geometry types.
    
    This enum defines the different types of geometric shapes that can be stored in a vector layer.
    Each type represents a specific geometric structure supported by the Geobox API.
    """
    Point = "Point"
    """A single point in 2D space"""
    
    MultiPoint = "MultiPoint"
    """A collection of points"""
    
    Polygon = "Polygon"
    """A closed shape defined by a sequence of points"""

    Polyline = "Polyline"
    """A polyline geometry"""


# Layer Enums
class FeatureType(Enum):
    """
    Enumeration of supported feature geometry types.
    
    This enum defines the different types of geometric shapes that can be stored in a feature.
    Each type represents a specific geometric structure supported by the Geobox API.
    """
    Point = "Point"
    """A single point in 2D space"""

    MultiPoint = "MultiPoint"
    """A collection of points"""
    
    LineString = "LineString"
    """A sequence of points forming a line"""

    MultiLineString = "MultiLineString"
    """A collection of lines"""
    
    Polygon = "Polygon"
    """A closed shape defined by a sequence of points"""
    
    MultiPolygon = "MultiPolygon"
    """A collection of polygons"""


# Task Enums
class TaskStatus(Enum):
    """
    Enumeration of possible task statuses.
    """
    PENDING = "PENDING"
    """
    The task is pending.
    """
    PROGRESS = "PROGRESS"
    """
    The task is in progress.
    """
    SUCCESS = "SUCCESS"
    """
    The task is successful.
    """
    FAILURE = "FAILURE"
    """
    The task is failed.
    """
    ABORTED = "ABORTED"
    """
    The task is aborted.
    """

class QueryGeometryType(Enum):
    """
    Enumeration of possible query geometry types.
    """
    POINT = "Point"
    """
    A point geometry.
    """
    MULTIPOINT = "Multipoint"
    """
    A multipoint geometry.
    """
    POLYLINE = "Polyline"
    """
    A polyline geometry.
    """
    POLYGON = "Polygon"
    """
    A polygon geometry.
    """

class QueryParamType(Enum):
    """
    Enumeration of possible query parameter types.
    """
    LAYER = "Layer"
    """
    A layer parameter.
    """
    ATTRIBUTE = "Attribute"
    """
    A Attribute parameter.
    """
    FLOAT = "Float"
    """
    A Float parameter.
    """
    INTEGER = "Integer"
    """
    A Integer parameter.
    """
    TEXT = "Text"
    """
    A Text parameter.
    """
    BOOLEAN = "Boolean"
    """
    A Boolean parameter.
    """


class UserRole(Enum):
    SYSTEM_ADMIN = "System Admin"

    ACCOUNT_ADMIN = "Account Admin"

    PUBLISHER = "Publisher"

    EDITOR = "Editor"

    VIEWER = "Viewer"


class UserStatus(Enum):
    NEW = "New"

    PENDING = "Pending"

    ACTIVE = "Active"

    DISABLED = "Disabled"


class MaxLogPolicy(Enum):
    OverwriteFirstLog = "OverwriteFirstLog"

    IgnoreLogging = "IgnoreLogging"


class InvalidDataPolicy(Enum):
    ReportToAdmin = "ReportToAdmin"

    LogOnly = "LogOnly"


class LoginFailurePolicy(Enum):
    BlockIPTemporarily = "BlockIPTemporarily"

    DisableAccount = "DisableAccount"


class MaxConcurrentSessionPolicy(Enum):
    CloseLastSession = "CloseLastSession"

    CloseMostInactiveSession = "CloseMostInactiveSession"

    PreventNewSession = "PreventNewSession"


class RoutingGeometryType(Enum):
    geojson = "geojson"

    polyline = "polyline"

    polyline6= "polyline6"


class RoutingOverviewLevel(Enum):
    Full = 'full'

    Simplified = 'simplified'


class QueryResultType(Enum):
    metadata = "metadata"

    data = "data"

    both = "both"


class FileOutputFormat(Enum):
    Shapefile = "Shapefile"

    GPKG = "GPKG"

    GeoJSON = "GeoJSON"

    CSV = "CSV"

    KML = "KML"

    DXF = "DXF"


class UsageScale(Enum):
    Hour = 'hour'

    Day = 'day'

    Month = 'month'


class UsageParam(Enum):
    Traffict = 'traffic'

    Calls = 'calls'


class AnalysisDataType(Enum):
    uint8 = 'uint8'
    uint16 = 'uint16'
    int16 = 'int16'
    uint32 = 'uint32'
    int32 = 'int32'
    float32 = 'float32'
    float64 = 'float64'


class PolygonizeConnectivity(Enum):
    connected_8 = 8
    connected_4 = 4


class AnalysisResampleMethod(Enum):
    near = 'near'
    bilinear = 'bilinear'
    cubic = 'cubic'
    cubicspline = 'cubicspline'
    lanczos = 'lanczos'
    average = 'average'
    mode = 'mode'
    max = 'max'
    min = 'min'
    med = 'med'
    q1 = 'q1'
    q3 = 'q3'
    rms = 'rms'
    sum = 'sum'


class SlopeUnit(Enum):
    degree = 'degree'
    percent = 'percent'


class AnalysisAlgorithm(Enum):
    Horn = 'Horn'
    ZevenbergenThorne = 'ZevenbergenThorne'


class RangeBound(Enum):
    left = 'left'
    right = 'right'
    both = 'both'
    neither = 'neither'


class DistanceUnit(Enum):
    GEO = 'GEO'
    PIXEL = 'PIXEL'


class GroupByAggFunction(Enum):
    COUNT = 'count'
    SUM = 'sum'
    MIN = 'min'
    MAX = 'max'
    AVG = 'avg'


class NetworkTraceDirection(Enum):
    UP = "up"
    DOWN = "down"


class SpatialAggFunction(Enum):
    COLLECT = 'collect'
    UNION = 'union'
    EXTENT = 'extent'
    MAKELINE = 'makeline'


class SpatialPredicate(Enum):
    INTERSECT = 'Intersect'
    CONTAIN = 'Contain'
    CROSS = 'Cross'
    EQUAL = 'Equal'
    OVERLAP = 'Overlap'
    TOUCH = 'Touch'
    WITHIN = 'Within'


class TableExportFormat(Enum):
    CSV = 'CSV'


"""
Geobox SDK

This package provides classes and functions to interact with the Geobox API.

Modules:
    api (module): Contains the Api class for making API requests.
    vectors (module): Contains classes for interacting with vector layers, fields, and features.
    exception (module): Contains custom exception classes for error handling.

Classes:
    Api: A class to interact with the GeoBox API.
    VectorLayers: A class to interact with vector layers in the GeoBox API.
    Fields: A class to interact with fields in a vector layer.
    Features: A class to interact with features in a vector layer.
    GeoboxError: Base class for all exceptions raised by the GeoBox SDK.
    AuthenticationError: Raised when there is an authentication error.
    AuthorizationError: Raised when there is an authorization error.
    ApiRequestError: Raised when there is an error with the API request.
    NotFoundError: Raised when a requested resource is not found.
    ValidationError: Raised when there is a validation error.
    ServerError: Raised when there is a server error.
"""
import logging, os, dotenv
from .api import GeoboxClient

from .exception import (
    GeoboxError,
    AuthenticationError,
    AuthorizationError,
    ApiRequestError,
    NotFoundError,
    ValidationError,
    ServerError
)

dotenv.load_dotenv()
DEBUG = os.getenv('DEBUG', 'False').lower() == 'true'

logger = logging.getLogger(__name__)
if DEBUG:
    logging.basicConfig(
        level=logging.DEBUG,
        format="%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    logger.info("Logging is set to DEBUG level.")
else:
    logging.basicConfig(level=logging.WARNING)


__all__ = [
    'GeoboxClient',
    'GeoboxError',
    'AuthenticationError',
    'AuthorizationError',
    'ApiRequestError',
    'NotFoundError',
    'ValidationError',
    'ServerError'
]




from urllib.parse import urljoin
from typing import Dict, List, TYPE_CHECKING, Union

from ..utils import clean_data
from .base import AsyncBase
from .task import AsyncTask
from ..enums import QueryResultType, QueryGeometryType, QueryParamType

if TYPE_CHECKING:
    from . import AsyncGeoboxClient
    from .user import AsyncUser
    from ..api import GeoboxClient
    from ..query import Query


class AsyncQuery(AsyncBase):

    BASE_ENDPOINT: str = 'queries/'

    def __init__(self, 
        api: 'AsyncGeoboxClient', 
        uuid: str = None, 
        data: Dict = {}):
        """
        Constructs all the necessary attributes for the Query object.

        Args:
            api (AsyncGeoboxClient): The API instance.
            uuid (str): The UUID of the query.
            data (dict, optional): The data of the query.
        """
        self.result = {}
        self._system_query = False
        super().__init__(api, uuid=uuid, data=data)


    def _check_access(self) -> None:
        """
        Check if the query is a system query.

        Returns:
            None

        Raises:
            PermissionError: If the query is a read-only system query.
        """
        if self._system_query:
            raise PermissionError("Cannot modify system queries - they are read-only")


    @property
    def sql(self) -> str:
        """
        Get the SQL of the query.
        
        Returns:
            str: The SQL of the query.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.query import AsyncQuery
            >>> async with AsyncGeoboxClient() as client:
            >>>     query = await AsyncQuery.get_query(client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     query.sql
            'SELECT * FROM some_layer'
        """
        return self.data['sql']
    

    @sql.setter
    def sql(self, value: str) -> None:
        """
        Set the SQL of the query.

        Args:
            value (str): The SQL of the query.

        Returns:
            None

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.query import AsyncQuery
            >>> async with AsyncGeoboxClient() as client:
            >>>     query = await AsyncQuery.get_query(client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     query.sql = 'SELECT * FROM some_layer'
            >>>     await query.save()
        """
        self.data['sql'] = value


    @property
    def params(self) -> List[Dict]:
        """
        Get the parameters of the query.

        Returns:
            List[Dict]: The parameters of the query.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.query import AsyncQuery
            >>> async with AsyncGeoboxClient() as client:
            >>>     query = await AsyncQuery.get_query(client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     query.params
            [{'name': 'layer', 'value': '12345678-1234-5678-1234-567812345678', 'type': 'Layer'}]
        """
        if not isinstance(self.data.get('params'), list):
            self.data['params'] = []

        return self.data['params']


    @params.setter
    def params(self, value: Dict) -> None:
        """
        Set the parameters of the query.

        Args:
            value (Dict): The parameters of the query.

        Returns:
            None

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.query import AsyncQuery
            >>> async with AsyncGeoboxClient() as client:
            >>>     query = await AsyncQuery.get_query(client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     query.params = [{'name': 'layer', 'value': '12345678-1234-5678-1234-567812345678', 'type': 'Layer'}]
            >>>     await query.save()
        """
        if not isinstance(self.data.get('params'), list):
            self.data['params'] = []
        
        self.data['params'] = value


    @classmethod
    async def get_queries(cls, api: 'AsyncGeoboxClient', **kwargs) -> Union[List['AsyncQuery'], int]:
        """
        [async] Get Queries

        Args:
            api (AsyncGeoboxClient): The AsyncGeoboxClient instance for making requests.

        Keyword Args:
            q (str): query filter based on OGC CQL standard. e.g. "field1 LIKE '%GIS%' AND created_at > '2021-01-01'"
            search (str): search term for keyword-based searching among search_fields or all textual fields if search_fields does not have value. NOTE: if q param is defined this param will be ignored.
            search_fields (str): comma separated list of fields for searching
            order_by (str): comma separated list of fields for sorting results [field1 A|D, field2 A|D, …]. e.g. name A, type D. NOTE: "A" denotes ascending order and "D" denotes descending order.
            return_count (bool): Whether to return total count. default is False.
            skip (int): Number of queries to skip. default is 0.
            limit(int): Maximum number of queries to return. default is 10.
            user_id (int): Specific user. privileges required.
            shared (bool): Whether to return shared queries. default is False.

        Returns:
            List[AsyncQuery] | int: list of queries or the number of queries.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.query import AsyncQuery
            >>> async with AsyncGeoboxClient() as client:
            >>>     queries = await AsyncQuery.get_queries(client)
            or  
            >>>     queries = await client.get_queries()
        """
        params = {
            'f': 'json',
            'q': kwargs.get('q'),
            'search': kwargs.get('search'),
            'search_field': kwargs.get('search_field'),
            'order_by': kwargs.get('order_by'),
            'return_count': kwargs.get('return_count', False),
            'skip': kwargs.get('skip', 0),
            'limit': kwargs.get('limit', 10),
            'user_id': kwargs.get('user_id'),
            'shared': kwargs.get('shared', False)
        }
        return await super()._get_list(api, cls.BASE_ENDPOINT, params, factory_func=lambda api, item: AsyncQuery(api, item['uuid'], item))


    @classmethod
    async def create_query(cls, api: 'AsyncGeoboxClient', name: str, display_name: str = None, description:str = None, sql: str = None, params: List = None) -> 'AsyncQuery':
        """
        [async] Creates a new query.

        Args:
            api (AsyncGeoboxClient): The AsyncGeoboxClient instance for making requests.
            name (str): The name of the query.
            display_name (str, optional): The display name of the query.
            description (str, optional): The description of the query.
            sql (str, optional): The SQL statement for the query.
            params (list, optional): The parameters for the SQL statement.

        Returns:
            AsyncQuery: The created query instance.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.query import AsyncQuery
            >>> async with AsyncGeoboxClient() as client:
            >>>     query = await AsyncQuery.create_query(client, name='query_name', display_name='Query Name', sql='SELECT * FROM some_layer')
            or  
            >>>     query = await client.create_query(name='query_name', display_name='Query Name', sql='SELECT * FROM some_layer')
        """
        data = {
            "name": name,
            "display_name": display_name,
            "description": description,
            "sql": sql,
            "params": params
        }
        return await super()._create(api, cls.BASE_ENDPOINT, data, factory_func=lambda api, item: AsyncQuery(api, item['uuid'], item))


    @classmethod
    async def get_query(cls, api: 'AsyncGeoboxClient', uuid: str, user_id: int = None) -> 'AsyncQuery':
        """
        [async] Retrieves a query by its UUID.

        Args:
            api (AsyncGeoboxClient): The AsyncGeoboxClient instance for making requests.
            uuid (str): The UUID of the query.
            user_id (int, optional): specific user ID. privileges required.

        Returns:
            AsyncQuery: The retrieved query instance.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.query import AsyncQuery
            >>> async with AsyncGeoboxClient() as client:
            >>>     query = await AsyncQuery.get_query(client, uuid="12345678-1234-5678-1234-567812345678")
            or  
            >>>     query = await client.get_query(uuid="12345678-1234-5678-1234-567812345678")
        """
        params = {
            'f': 'json',
            'user_id': user_id
        }
        return await super()._get_detail(api, cls.BASE_ENDPOINT, uuid, params, factory_func=lambda api, item: AsyncQuery(api, item['uuid'], item))


    @classmethod
    async def get_query_by_name(cls, api: 'AsyncGeoboxClient', name: str, user_id: int = None) -> Union['AsyncQuery', None]:
        """
        [async] Get a query by name

        Args:
            api (AsyncGeoboxClient): The AsyncGeoboxClient instance for making requests.
            name (str): the name of the query to get
            user_id (int, optional): specific user. privileges required.

        Returns:
            AsyncQuery | None: returns the query if a query matches the given name, else None

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.query import AsyncQuery
            >>> async with AsyncGeoboxClient() as client:
            >>>     query = await AsyncQuery.get_query_by_name(client, name='test')
            or  
            >>>     query = await client.get_query_by_name(name='test')
        """
        queries = await cls.get_queries(api, q=f"name = '{name}'", user_id=user_id)
        if queries and queries[0].name == name:
            return queries[0]
        else:
            return None


    @classmethod
    async def get_system_queries(cls, api: 'AsyncGeoboxClient', **kwargs) -> List['AsyncQuery']:
        """
        [async] Returns the system queries as a list of Query objects.

        Args:
            api (AsyncGeoboxClient): The AsyncGeoboxClient instance for making requests.

        Keyword Args:
            q (str): query filter based on OGC CQL standard. e.g. "field1 LIKE '%GIS%' AND created_at > '2021-01-01'".
            search (str): search term for keyword-based searching among search_fields or all textual fields if search_fields does not have value. NOTE: if q param is defined this param will be ignored.
            search_fields (str): comma separated list of fields for searching.
            order_by (str): comma separated list of fields for sorting results [field1 A|D, field2 A|D, …]. e.g. name A, type D. NOTE: "A" denotes ascending order and "D" denotes descending order.
            return_count (bool): whether to return the total count of queries. default is False.
            skip (int): number of queries to skip. minimum is 0. default is 0.
            limit (int): number of queries to return. minimum is 1. default is 100.
            user_id (int): specific user. privileges required.
            shared (bool): whether to return shared queries. default is False.
        
        Returns:
            List[AsyncQuery]: list of system queries.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.query import AsyncQuery
            >>> async with AsyncGeoboxClient() as client:
            >>>     queries = await AsyncQuery.get_system_queries(client)
            or  
            >>>     queries = await client.get_system_queries()
        """
        params = {
            'f': 'json',
            'q': kwargs.get('q'),
            'search': kwargs.get('search'),
            'search_fields': kwargs.get('search_fields'),
            'order_by': kwargs.get('order_by'),
            'return_count': kwargs.get('return_count', False),
            'skip': kwargs.get('skip', 0),
            'limit': kwargs.get('limit', 100),
            'user_id': kwargs.get('user_id'),
            'shared': kwargs.get('shared', False)            
        }
        endpoint = urljoin(cls.BASE_ENDPOINT, 'systemQueries/')
        def factory_func(api, item):
            query = AsyncQuery(api, item['uuid'], item)
            query._system_query = True
            return query
            
        return await super()._get_list(api, endpoint, params, factory_func=factory_func)

    
    def add_param(self, name: str, value: str, type: 'QueryParamType', default_value: str = None, Domain: Dict = None) -> None:
        """
        Add a parameter to the query parameters.
        
        Args:
            name (str): The name of the parameter.
            value (str): The value of the parameter.
            type (str): The type of the parameter (default: 'Layer').
            default_value (str, optional): The default value for the parameter.
            Domain (Dict, optional): Domain information for the parameter.
            
        Returns:
            None
            
        Raises:
            PermissionError: If the query is a read-only system query.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.query import AsyncQuery
            >>> async with AsyncGeoboxClient() as client:
            >>>     query = await AsyncQuery.get_query(client, uuid="12345678-1234-5678-1234-567812345678")
            or  
            >>>     query = await client.get_query(uuid="12345678-1234-5678-1234-567812345678")
            >>>     query.add_param(name='param_name', value='param_value', type=QueryParamType.LAYER)
            >>>     await query.save()
        """
        self.params.append({
            'name': name,
            'value': value,
            'type': type.value,
            'default_value': default_value,
            'Domain': Domain
        })


    def remove_param(self, name: str) -> None:
        """
        Remove a parameter from the query parameters by name.
        
        Args:
            name (str): The name of the parameter to remove.
            
        Returns:
            None
            
        Raises:
            ValueError: If the parameter is not found in query parameters.
            PermissionError: If the query is a read-only system query.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.query import AsyncQuery
            >>> async with AsyncGeoboxClient() as client:
            >>>     query = await AsyncQuery.get_query(client, uuid="12345678-1234-5678-1234-567812345678")
            or  
            >>>     query = await client.get_query(uuid="12345678-1234-5678-1234-567812345678")
            >>>     query.remove_param(name='param_name')
            >>>     await quary.save()
        """
        for i, param in enumerate(self.params):
            if param.get('name') == name:
                self.params.pop(i)
                return
        
        raise ValueError(f"Parameter with name '{name}' not found in query parameters")


    async def execute(self, 
                f: str = "json",
                result_type: QueryResultType = QueryResultType.both, 
                return_count: bool = None,
                out_srid: int = None, 
                quant_factor: int = 1000000, 
                bbox_srid: int = None, 
                skip: int = None, 
                limit: int = None,
                skip_geometry: bool = False) -> Union[Dict, int]:
        """
        [async] Executes a query with the given SQL statement and parameters.

        Args:
            f (str): the output format of the executed query. options are: json, topojson. default is json.
            result_type (QueryResultType, optional): The type of result to return (default is "both").
            return_count (bool, optional): Whether to return the count of results.
            out_srid (int, optional): The output spatial reference ID.
            quant_factor (int, optional): The quantization factor (default is 1000000).
            bbox_srid (int, optional): The bounding box spatial reference ID.
            skip (int, optional): The number of results to skip.
            limit (int, optional): The maximum number of results to return.
            skip_geometry (bool): Whether to skip the geometry part of the features or not. default is False.

        Returns:
            Dict | int: The result of the query execution or the count number of the result

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.query import AsyncQuery
            >>> async with AsyncGeoboxClient() as client:
            >>>     query = await AsyncQuery.get_query(client, uuid="12345678-1234-5678-1234-567812345678")
            or  
            >>>     query = await client.get_query(uuid="12345678-1234-5678-1234-567812345678")
            >>>     await query.execute(f='json')
        """
        if not self.sql:
            raise ValueError('"sql" parameter is required for this action!') 
        if not self.params:
            raise ValueError('"params" parameter is required for this action!')

        data = clean_data({
            "f": f if f in ['json', 'topojson'] else None,
            "sql": self.sql,
            "params": self.params,
            "result_type": result_type.value,
            "return_count": return_count,
            "out_srid": out_srid,
            "quant_factor": quant_factor,
            "bbox_srid": bbox_srid,
            "skip": skip,
            "limit": limit,
            "skip_geometry": skip_geometry
        })

        endpoint = urljoin(self.BASE_ENDPOINT, 'exec/')
        self.result = await self.api.post(endpoint, data)
        return self.result


    async def update(self, **kwargs) -> Dict:
        """
        [async] Updates the query with new data.

        Keyword Args:
            name (str): The new name of the query.
            display_name (str): The new display name of the query.
            sql (str): The new SQL statement for the query.
            params (list): The new parameters for the SQL statement.

        Returns:
            Dict: The updated query data.
            
        Raises:
            PermissionError: If the query is a read-only system query.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.query import AsyncQuery
            >>> async with AsyncGeoboxClient() as client:
            >>>     query = await AsyncQuery.get_query(client, uuid="12345678-1234-5678-1234-567812345678")
            or  
            >>>     query = await client.get_query(uuid="12345678-1234-5678-1234-567812345678")
            >>>     await query.update(name='new_name')
        """
        self._check_access()
            
        data = {
            "name": kwargs.get('name'),
            "display_name": kwargs.get('display_name'),
            "sql": kwargs.get('sql'),
            "params": kwargs.get('params')
        }
        return await super()._update(self.endpoint, data)


    async def save(self) -> None:
        """
        [async] Save the query. Creates a new query if query uuid is None, updates existing query otherwise.

        Returns:
            None

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.query import AsyncQuery
            >>> async with AsyncGeoboxClient() as client:
            >>>     query = await AsyncQuery.get_query(client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     await query.save()
        """
        self.params = [item for item in self.params if item.get('value')]

        try:
            if self.__getattr__('uuid'):
                await self.update(name=self.data['name'], display_name=self.data['display_name'], sql=self.sql, params=self.params)
        except AttributeError:
            response = await self.api.post(self.BASE_ENDPOINT, self.data)
            self.endpoint = urljoin(self.BASE_ENDPOINT, f'{response["uuid"]}/')        
            self.data.update(response)


    async def delete(self) -> str:
        """
        [async] Deletes a query.

        Returns:
            str: The response from the API.
            
        Raises:
            PermissionError: If the query is a read-only system query

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.query import AsyncQuery
            >>> async with AsyncGeoboxClient() as client:
            >>>     query = await AsyncQuery.get_query(client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     await query.delete()
        """
        self._check_access()
        await super()._delete(self.endpoint)


    async def share(self, users: List['AsyncUser']) -> None:
        """
        [async] Shares the query with specified users.

        Args:
            users (List[AsyncUser]): The list of user objects to share the query with.

        Returns:
            None
            
        Raises:
            PermissionError: If the query is a read-only system query.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.query import AsyncQuery
            >>> async with AsyncGeoboxClient() as client:
            >>>     query = await AsyncQuery.get_query(client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     users = await client.search_users(search="John")
            >>>     await query.share(users=users)
        """
        self._check_access()
        await super()._share(self.endpoint, users)
    

    async def unshare(self, users: List['AsyncUser']) -> None:
        """
        [async] Unshares the query with specified users.

        Args:
            users (List[AsyncUser]): The list of user objects to unshare the query with.

        Returns:
            None
            
        Raises:
            PermissionError: If the query is a read-only system query.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.query import AsyncQuery
            >>> async with AsyncGeoboxClient() as client:
            >>>     query = await AsyncQuery.get_query(client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     users = await client.search_users(search="John")
            >>>     await query.unshare(users=users)
        """
        self._check_access()
        await super()._unshare(self.endpoint, users)


    async def get_shared_users(self, search: str = None, skip: int = 0, limit: int = 10) -> List['AsyncUser']:
        """
        [async] Retrieves the list of users the query is shared with.

        Args:
            search (str, optional): the search query.
            skip (int, optional): The number of users to skip.
            limit (int, optional): The maximum number of users to retrieve.

        Returns:
            List[AsyncUser]: The list of shared users.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.query import AsyncQuery
            >>> async with AsyncGeoboxClient() as client:
            >>>     query = await AsyncQuery.get_query(client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     users = await client.search_users(search="John")
            >>>     await query.get_shared_users(search='John', skip=0, limit=10)
        """
        self._check_access()
        params = {
            'search': search,
            'skip': skip,
            'limit': limit
        }
        return await super()._get_shared_users(self.endpoint, params)


    @property
    def thumbnail(self) -> str:
        """
        Retrieves the thumbnail URL for the query.            

        Returns:
            str: The thumbnail URL.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.query import AsyncQuery
            >>> async with AsyncGeoboxClient() as client:
            >>>     query = await AsyncQuery.get_query(client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     query.thumbnail
        """
        return super()._thumbnail()


    async def save_as_layer(self, layer_name: str, layer_type: 'QueryGeometryType' = None) -> 'AsyncTask':
        """
        [async] Saves the query as a new layer.

        Args:
            layer_name (str): The name of the new layer.
            layer_type (QueryGeometryType, optional): The type of the new layer.

        Returns:
            Task: The response task object.
            
        Raises:
            PermissionError: If the query is a read-only system query.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.query import AsyncQuery
            >>> async with AsyncGeoboxClient() as client:
            >>>     query = await AsyncQuery.get_query(client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     await query.save_as_layer(layer_name='test')
        """
        params = [{
            "name": item.get('name'),
            "type": item.get('type'),
            "value": item.get('default_value') if not item.get('value') else item.get('value')
        } for item in self.params]

        data = clean_data({
            "sql": self.sql,
            "params": params,
            "layer_name": layer_name,
            "layer_type": layer_type.value if layer_type else None
        })

        endpoint = urljoin(self.BASE_ENDPOINT, 'saveAsLayer/')
        response = await self.api.post(endpoint, data)
        task = await AsyncTask.get_task(self.api, response.get('task_id'))
        return task


    def to_sync(self, sync_client: 'GeoboxClient') -> 'Query':
        """
        Switch to sync version of the query instance to have access to the sync methods

        Args:
            sync_client (GeoboxClient): The sync version of the GeoboxClient instance for making requests.

        Returns:
            Query: the sync instance of the query.

        Example:
            >>> from geobox import Geoboxclient
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.query import AsyncQuery
            >>> client = GeoboxClient()
            >>> async with AsyncGeoboxClient() as async_client:
            >>>     query = await AsyncQuery.get_query(async_client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     sync_query = query.to_sync(client)
        """
        from ..query import Query

        return Query(api=sync_client, uuid=self.uuid, data=self.data)


from typing import List, Dict, Optional, TYPE_CHECKING, Union

from .base import AsyncBase

if TYPE_CHECKING:
    from . import AsyncGeoboxClient
    from .user import AsyncUser
    from ..api import GeoboxClient
    from ..dashboard import Dashboard


class AsyncDashboard(AsyncBase):

    BASE_ENDPOINT = 'dashboards/'

    def __init__(self, 
        api: 'AsyncGeoboxClient', 
        uuid: str,
        data: Optional[Dict] = {}):
        """
        Initialize a Dashboard instance.

        Args:
            api (AsyncGeoboxClient): The AsyncGeoboxClient instance for making requests.
            uuid (str): The unique identifier for the Dashboard.
            data (Dict, optional): The data of the Dashboard.
        """
        super().__init__(api, uuid=uuid, data=data)


    @classmethod
    async def get_dashboards(cls, api: 'AsyncGeoboxClient', **kwargs) -> Union[List['AsyncDashboard'], int]:
        """
        [async] Get list of Dashboards

        Args:
            api (AsyncGeoboxClient): The AsyncGeoboxClient instance for making requests.

        Keyword Args:
            q (str): query filter based on OGC CQL standard. e.g. "field1 LIKE '%GIS%' AND created_at > '2021-01-01'"
            search (str): search term for keyword-based searching among search_fields or all textual fields if search_fields does not have value. NOTE: if q param is defined this param will be ignored.
            search_fields (str): comma separated list of fields for searching.
            order_by (str): comma separated list of fields for sorting results [field1 A|D, field2 A|D, …]. e.g. name A, type D. NOTE: "A" denotes ascending order and "D" denotes descending order.
            return_count (bool): Whether to return total count. default is False.
            skip (int): Number of items to skip. default is 0.
            limit (int): Number of items to return. default is 10.
            user_id (int): Specific user. privileges required.
            shared (bool): Whether to return shared Dashboards. default is False.

        Returns:
            List[AsyncDashboard] | int: A list of Dashboard instances or the total number of Dashboards.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.dashboard import AsyncDashboard
            >>> async with AsyncGeoboxClient() as client:
            >>>     dashboards = await AsyncDashboard.get_dashboards(client)
            or  
            >>>     dashboards = await client.get_dashboards()
        """
        params = {
           'f': 'json',
           'q': kwargs.get('q'),
           'search': kwargs.get('search'),
           'search_fields': kwargs.get('search_fields'),
           'order_by': kwargs.get('order_by'),
           'return_count': kwargs.get('return_count', False),
           'skip': kwargs.get('skip', 0),
           'limit': kwargs.get('limit', 10),
           'user_id': kwargs.get('user_id'),
           'shared': kwargs.get('shared', False)
        }
        return await super()._get_list(api, cls.BASE_ENDPOINT, params, factory_func=lambda api, item: AsyncDashboard(api, item['uuid'], item))
    

    @classmethod
    async def create_dashboard(cls, 
        api: 'AsyncGeoboxClient', 
        name: str, 
        display_name: str = None, 
        description: str = None, 
        settings: Dict = {}, 
        thumbnail: str = None, 
        user_id: int = None) -> 'AsyncDashboard':
        """
        [async] Create a new Dashboard.

        Args:
            api (AsyncGeoboxClient): The AsyncGeoboxClient instance for making requests.
            name (str): The name of the Dashboard.
            display_name (str, optional): The display name of the Dashboard.
            description (str, optional): The description of the Dashboard.
            settings (Dict, optional): The settings of the sceDashboarde.
            thumbnail (str, optional): The thumbnail of the Dashboard.
            user_id (int, optional): Specific user. privileges required.

        Returns:
            AsyncDashboard: The newly created Dashboard instance.

        Raises:
            ValidationError: If the Dashboard data is invalid.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.dashboard import AsyncDashboard
            >>> async with AsyncGeoboxClient() as client:
            >>>     dashboard = await AsyncDashboard.create_dashboard(client, name="my_dashboard")
            or  
            >>>     dashboard = await client.create_dashboard(name="my_dashboard")
        """
        data = {
            "name": name,
            "display_name": display_name,
            "description": description,
            "settings": settings,
            "thumbnail": thumbnail,
            "user_id": user_id,
        }
        return await super()._create(api, cls.BASE_ENDPOINT, data, factory_func=lambda api, item: AsyncDashboard(api, item['uuid'], item))

    
    @classmethod
    async def get_dashboard(cls, api: 'AsyncGeoboxClient', uuid: str, user_id: int = None) -> 'AsyncDashboard':
        """
        [async] Get a Dashboard by its UUID.

        Args:
            api (AsyncGeoboxClient): The AsyncGeoboxClient instance for making requests.
            uuid (str): The UUID of the Dashboard to get.
            user_id (int, optional): Specific user. privileges required.

        Returns:
            AsyncDashboard: The dashboard object.

        Raises:
            NotFoundError: If the Dashboard with the specified UUID is not found.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.dashboard import AsyncDashboard
            >>> async with AsyncGeoboxClient() as client:
            >>>     dashboard = await AsyncDashboard.get_dashboard(client, uuid="12345678-1234-5678-1234-567812345678")
            or  
            >>>     dashboard = await client.get_dashboard(uuid="12345678-1234-5678-1234-567812345678")
        """
        params = {
            'f': 'json',
            'user_id': user_id,
        }
        return await super()._get_detail(api, cls.BASE_ENDPOINT, uuid, params, factory_func=lambda api, item: AsyncDashboard(api, item['uuid'], item))


    @classmethod
    async def get_dashboard_by_name(cls, api: 'AsyncGeoboxClient', name: str, user_id: int = None) -> Union['AsyncDashboard', None]:
        """
        [async] Get a dashboard by name

        Args:
            api (AsyncGeoboxClient): The AsyncGeoboxClient instance for making requests.
            name (str): the name of the dashboard to get
            user_id (int, optional): specific user. privileges required.

        Returns:
            AsyncDashboard | None: returns the dashboard if a dashboard matches the given name, else None

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.dashboard import AsyncDashboard
            >>> async with AsyncGeoboxClient() as client:
            >>>     dashboard = await AsyncDashboard.get_dashboard_by_name(client, name='test')
            or  
            >>>     dashboard = await client.get_dashboard_by_name(name='test')
        """
        dashboards = await cls.get_dashboards(api, q=f"name = '{name}'", user_id=user_id)
        if dashboards and dashboards[0].name == name:
            return dashboards[0]
        else:
            return None


    async def update(self, **kwargs) -> Dict:
        """
        [async] Update the Dashboard

        Keyword Args:
            name (str): The name of the Dashboard.
            display_name (str): The display name of the Dashboard.
            description (str): The description of the Dashboard.
            settings (Dict): The settings of the Dashboard.
            thumbnail (str): The thumbnail of the Dashboard.

        Returns:
            Dict: The updated Dashboard data.

        Raises:
            ValidationError: If the Dashboard data is invalid.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.dashboard import AsyncDashboard
            >>> async with AsyncGeoboxClient() as client:
            >>>     dashboard = await AsyncDashboard.get_dashboard(client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     await dashboard.update(display_name="New Display Name")
        """
        data = {
            "name": kwargs.get('name'),
            "display_name": kwargs.get('display_name'),   
            "description": kwargs.get('description'),
            "settings": kwargs.get('settings'),
            "thumbnail": kwargs.get('thumbnail')
        }
        return await super()._update(self.endpoint, data)
    

    async def delete(self) -> None:
        """
        [async] Delete the dashboard.

        Returns:
            None

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.dashboard import AsyncDashboard
            >>> async with AsyncGeoboxClient() as client:
            >>>     dashboard = await AsyncDashboard.get_dashboard(client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     await dashboard.delete()
        """        
        await super()._delete(self.endpoint)


    @property
    def thumbnail(self) -> str:
        """
        Get the thumbnail URL of the dashboard.

        Returns:
            str: The thumbnail of the dashboard.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.dashboard import AsyncDashboard
            >>> async with AsyncGeoboxClient() as client:
            >>>     dashboard = await AsyncDashboard.get_dashboard(client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     dashboard.thumbnail
        """
        return super()._thumbnail()
    

    async def share(self, users: List['AsyncUser']) -> None:
        """
        [async] Shares the Dashboard with specified users.

        Args:
            users (List[AsyncUser]): The list of user objects to share the Dashboard with.

        Returns:
            None

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.dashboard import AsyncDashboard
            >>> async with AsyncGeoboxClient() as client:
            >>>     dashboard = await AsyncDashboard.get_dashboard(client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     users = await client.search_users(search='John')
            >>>     await dashboard.share(users=users)
        """
        await super()._share(self.endpoint, users)
    

    async def unshare(self, users: List['AsyncUser']) -> None:
        """
        [async] Unshares the Dashboard with specified users.

        Args:
            users (List[AsyncUser]): The list of user objects to unshare the Dashboard with.

        Returns:
            None

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.dashboard import AsyncDashboard
            >>> async with AsyncGeoboxClient() as client:
            >>>     dashboard = await AsyncDashboard.get_dashboard(client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     users = await client.search_users(search='John')
            >>>     await dashboard.unshare(users=users)
        """
        await super()._unshare(self.endpoint, users)


    async def get_shared_users(self, search: str = None, skip: int = 0, limit: int = 10) -> List['AsyncUser']:
        """
        [async] Retrieves the list of users the dashboard is shared with.

        Args:
            search (str, optional): The search query.
            skip (int, optional): The number of users to skip.
            limit (int, optional): The maximum number of users to retrieve.

        Returns:
            List[User]: The list of shared users.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.dashboard import AsyncDashboard
            >>> async with AsyncGeoboxClient() as client:
            >>>     dashboard = await AsyncDashboard.get_dashboard(client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     await dashboard.get_shared_users(search='John', skip=0, limit=10)
        """
        params = {
            'search': search,
            'skip': skip,
            'limit': limit
        }
        return await super()._get_shared_users(self.endpoint, params)


    def to_sync(self, sync_client: 'GeoboxClient') -> 'Dashboard':
        """
        Switch to sync version of the dashboard instance to have access to the sync methods

        Args:
            sync_client (GeoboxClient): The sync version of the GeoboxClient instance for making requests.

        Returns:
            Dashboard: the sync instance of the dashboard.

        Example:
            >>> from geobox import Geoboxclient
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.dashboard import AsyncDashboard
            >>> client = GeoboxClient()
            >>> async with AsyncGeoboxClient() as async_client:
            >>>     dashboard = await AsyncDashboard.get_dashboard(async_client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     sync_dashboard = dashboard.to_sync(client)
        """
        from ..dashboard import Dashboard

        return Dashboard(api=sync_client, uuid=self.uuid, data=self.data)


import pytest
from unittest.mock import patch

from geobox.aio.table import AsyncTableField, AsyncTableRow
from geobox.enums import FieldType
from geobox.exception import NotFoundError
from geobox.file import File
from geobox.table import Table, TableField, TableRow
from geobox.task import Task
from geobox.user import User

def test_init(api, mock_table_data):
    table = Table(api, mock_table_data['uuid'], mock_table_data)
    assert table.name == mock_table_data['name']
    assert table.uuid == mock_table_data['uuid']
    assert table.data == mock_table_data
    assert table.endpoint == f'{Table.BASE_ENDPOINT}{table.uuid}/'


def test_get_tables(api, mock_table_data):
    api.get.return_value = [mock_table_data, mock_table_data]
    tables = Table.get_tables(api)
    api.get.assert_called_once_with(f'{Table.BASE_ENDPOINT}?f=json&include_settings=False&return_count=False&skip=0&limit=10&shared=False')
    assert len(tables) == 2
    assert type(tables[0]) == Table
    assert tables[0].data == mock_table_data


def test_create_table(api, mock_table_data, mock_table_field_data):
    api.post.return_value = mock_table_data
    table = Table.create_table(api,
        name=mock_table_data['name'],
        display_name=mock_table_data['display_name'],
        description=mock_table_data['description'],
        temporary=mock_table_data['temporary'],
        fields=[mock_table_field_data])
    api.post.assert_called_once_with(Table.BASE_ENDPOINT, {'name': 'test', 'temporary': False, 'fields': [mock_table_field_data]})
    assert type(table) == Table
    assert table.uuid == mock_table_data['uuid']
    assert table.data == mock_table_data


def test_get_table(api, mock_table_data):
    api.get.return_value = mock_table_data
    table = Table.get_table(api, uuid=mock_table_data['uuid'])
    api.get.assert_called_once_with(f"{Table.BASE_ENDPOINT}{mock_table_data['uuid']}/?f=json")
    assert type(table) == Table
    assert table.uuid == mock_table_data['uuid']
    assert table.data == mock_table_data


def test_get_table_by_name(api, mock_table_data):
    api.get.return_value = [mock_table_data]
    table = Table.get_table_by_name(api, name=mock_table_data['name'])
    api.get.assert_called_once_with(f"{Table.BASE_ENDPOINT}?f=json&include_settings=False&q=name+%3D+%27test%27&return_count=False&skip=0&limit=10&shared=False")
    assert type(table) == Table
    assert table.uuid == mock_table_data['uuid']
    assert table.data == mock_table_data
    # not found
    table = Table.get_table_by_name(api, name='not_found')
    assert table == None


def test_update(api, mock_table_data):
    table = Table(api, mock_table_data['uuid'], mock_table_data)
    updated_data = {
        **mock_table_data,
        'name': 'update',
        'display_name': 'update display name',
        'description': 'update description',
    }

    api.put.return_value = updated_data
    result = table.update(name='update', display_name='update display name', description='update description')
    api.put.assert_called_once_with(table.endpoint, {'name': 'update', 'display_name': 'update display name', 'description': 'update description'})
    assert updated_data == result
    assert table.data == updated_data


def test_delete(api, mock_table_data):
    table = Table(api, mock_table_data['uuid'], mock_table_data)
    endpoint = table.endpoint
    table.delete()
    api.delete.assert_called_once_with(endpoint)
    assert table.uuid is None
    assert table.endpoint is None


def test_get_settings(api, mock_table_data):
    table = Table(api, mock_table_data['uuid'], mock_table_data)
    api.get.return_value = mock_table_data['settings']
    settings = table.settings
    assert settings == mock_table_data['settings']
    api.get.assert_called_once_with(f"{table.endpoint}settings/?f=json")


def test_update_settings(api, mock_table_data):
    table = Table(api, mock_table_data['uuid'], mock_table_data)
    api.get.return_value = mock_table_data['settings']
    settings = table.settings
    settings['edit_settings']['allow_delete'] = False
    api.put.return_value = settings
    result = table.update_settings(settings)
    assert result == settings


def test_get_fields(api, mock_table_data, mock_table_field_data):
    table = Table(api, mock_table_data['uuid'], mock_table_data)
    api.get.return_value = [mock_table_field_data, mock_table_field_data]
    fields = table.get_fields()
    assert len(fields) == 2
    assert type(fields[0]) == TableField
    assert fields[0].data == mock_table_field_data


def test_get_field(api, mock_table_data, mock_table_field_data):
    table = Table(api, mock_table_data['uuid'], mock_table_data)
    api.get.return_value = [mock_table_field_data, mock_table_field_data]
    field = table.get_field(field_id=mock_table_field_data['id'])
    assert type(field) == TableField
    assert field.data == mock_table_field_data


def test_get_field_by_name(api, mock_table_data, mock_table_field_data):
    table = Table(api, mock_table_data['uuid'], mock_table_data)
    api.get.return_value = [mock_table_field_data, mock_table_field_data]
    field = table.get_field_by_name(mock_table_field_data['name'])
    assert type(field) == TableField
    assert field.data == mock_table_field_data


def test_get_field_not_found_error(api, mock_table_data, mock_table_field_data):
    table = Table(api, mock_table_data['uuid'], mock_table_data)
    api.get.return_value = [mock_table_field_data, mock_table_field_data]
    with pytest.raises(NotFoundError):
        table.get_field(10)

    with pytest.raises(NotFoundError):
        table.get_field_by_name('t')


def test_add_field(api, mock_table_data, mock_table_field_data):
    table = Table(api, mock_table_data['uuid'], mock_table_data)
    api.post.return_value = mock_table_field_data
    field = table.add_field(name='field1', data_type=FieldType.String)
    assert type(field) == TableField
    assert field.data == mock_table_field_data


def test_calculate_field(api, mock_table_data, mock_success_task_data):
    table = Table(api, mock_table_data['uuid'], mock_table_data)
    api.get.return_value = mock_success_task_data
    task = table.calculate_field(target_field='field1', expression='upper( field1  )')
    assert type(task) == Task
    api.post.assert_called_once_with(f"{table.endpoint}calculateField/", {'target_field': 'field1', 'expression': 'upper( field1  )', 'run_async': True}, is_json=False)


def test_calculate_field_sync(api, mock_table_data):
    table = Table(api, mock_table_data['uuid'], mock_table_data)
    api.post.return_value = {'result': 'value'}
    result = table.calculate_field(target_field='field1', expression='upper( field1  )', run_async=False)
    assert result == {'result': 'value'}
    api.post.assert_called_once_with(f"{table.endpoint}calculateField/", {'target_field': 'field1', 'expression': 'upper( field1  )', 'run_async': False}, is_json=False)


def test_get_rows(api, mock_table_data, mock_table_row_data):
    table = Table(api, mock_table_data['uuid'], mock_table_data)
    api.get.return_value = [mock_table_row_data, mock_table_row_data]
    rows = table.get_rows()
    assert len(rows) == 2
    assert type(rows[0]) == TableRow
    assert rows[0].data == mock_table_row_data
    api.get.assert_called_once_with(f"{table.endpoint}rows/?f=json&skip=0&limit=100&return_count=False")


def test_get_row(api, mock_table_data, mock_table_row_data):
    table = Table(api, mock_table_data['uuid'], mock_table_data)
    api.get.return_value = mock_table_row_data
    row = table.get_row(row_id=mock_table_row_data['id'])
    assert type(row) == TableRow
    assert row.data == mock_table_row_data
    api.get.assert_called_once_with(f"{row.endpoint}?f=json")


def test_create_row(api, mock_table_data, mock_table_row_data):
    table = Table(api, mock_table_data['uuid'], mock_table_data)
    api.post.return_value = mock_table_row_data
    row = table.create_row(field1='test')
    assert type(row) == TableRow
    assert row.data == mock_table_row_data
    api.post.assert_called_once_with(f"{table.endpoint}rows/", {'field1': 'test'})


def test_import_rows(api, mock_table_data, mock_file_data, mock_success_task_data):
    table = Table(api, mock_table_data['uuid'], mock_table_data)
    file = File(api, mock_file_data['uuid'], mock_file_data)
    api.post.return_value = {'task_id': mock_success_task_data['uuid']}
    api.get.return_value = mock_success_task_data
    task = table.import_rows(file)
    assert type(task) == Task
    assert task.data == mock_success_task_data
    api.post.assert_called_once_with(f"{table.endpoint}import-csv/", {'file_uuid': '2228a886-870e-4a0a-8e3b-6f51efee2eae', 'file_encoding': 'utf-8', 'input_dataset': 'world_boundaries.shp', 'delimiter': ',', 'has_header': True, 'report_errors': False, 'bulk_insert': True}, is_json=False)
    api.get.assert_called_once_with(task.endpoint)


def test_export_rows(api, mock_table_data, mock_success_task_data):
    table = Table(api, mock_table_data['uuid'], mock_table_data)
    api.post.return_value = {'task_id': mock_success_task_data['uuid']}
    api.get.return_value = mock_success_task_data
    task = table.export_rows('test')
    assert type(task) == Task
    assert task.data == mock_success_task_data
    api.post.assert_called_once_with(f"{table.endpoint}export/", {'out_filename': 'test', 'out_format': 'CSV', 'zipped': False, 'run_async': True}, is_json=False)
    api.get.assert_called_once_with(task.endpoint)


def test_export_rows_sync(api, mock_table_data):
    table = Table(api, mock_table_data['uuid'], mock_table_data)
    api.post.return_value = {'result': 'value'}
    result = table.export_rows('test', run_async=False)
    assert result == {'result': 'value'}
    api.post.assert_called_once_with(f"{table.endpoint}export/", {'out_filename': 'test', 'out_format': 'CSV', 'zipped': False, 'run_async': False}, is_json=False)


def test_row_repr(api, mock_table_data, mock_table_row_data):
    table = Table(api, mock_table_data['uuid'], mock_table_data)
    row = TableRow(table, mock_table_row_data)
    assert repr(row) == f"TableRow(id={row.id}, table_name={table.name})"


def test_row_update(api, mock_table_data, mock_table_row_data):
    table = Table(api, mock_table_data['uuid'], mock_table_data)
    row = TableRow(table, mock_table_row_data)
    updated_data = mock_table_row_data
    updated_data['properties'] = {
        'field1': 'updated row'
    }
    api.put.return_value = updated_data
    result = row.update(field1='updated row')
    assert row.properties['field1'] == 'updated row'
    api.put.assert_called_once_with(row.endpoint, {'id': 40, 'properties': {'field1': 'updated row'}})
    assert result == updated_data
    

def test_row_delete(api, mock_table_data, mock_table_row_data):
    table = Table(api, mock_table_data['uuid'], mock_table_data)
    row = TableRow(table, mock_table_row_data)
    endpoint = row.endpoint
    row.delete()
    api.delete.assert_called_once_with(endpoint)


def test_row_to_async(api, async_api, mock_table_data, mock_table_row_data):
    table = Table(api, uuid=mock_table_data['uuid'], data=mock_table_data)
    row = TableRow(table, mock_table_row_data)
    async_instance = row.to_async(async_api)
    assert async_instance.api == async_api
    assert type(async_instance) == AsyncTableRow


def test_field_data_type(api, mock_table_data, mock_table_field_data):
    table = Table(api, mock_table_data['uuid'], mock_table_data)
    field = TableField(table, data_type=FieldType.String, field_id=mock_table_field_data['id'], data=mock_table_field_data)
    assert field.datatype == FieldType.String


def test_field_init_datatype_error(api, mock_table_data, mock_table_field_data):
    table = Table(api, mock_table_data['uuid'], mock_table_data)
    with pytest.raises(ValueError):
        TableField(table, data_type='string', field_id=mock_table_field_data['id'], data=mock_table_field_data)


def test_field_repr(api, mock_table_data, mock_table_field_data):
    table = Table(api, mock_table_data['uuid'], mock_table_data)
    field = TableField(table, data_type=FieldType.String, field_id=mock_table_field_data['id'], data=mock_table_field_data)
    assert repr(field) == f'TableField(id={field.id}, name={field.name}, data_type={field.data_type})'


def test_field_get_domain(api, mock_table_data, mock_table_field_data):
    table = Table(api, mock_table_data['uuid'], mock_table_data)
    field = TableField(table, data_type=FieldType.String, field_id=mock_table_field_data['id'], data=mock_table_field_data)
    assert field.domain == mock_table_field_data['domain']
    

def test_field_set_domain(api, mock_table_data, mock_table_field_data):
    table = Table(api, mock_table_data['uuid'], mock_table_data)
    field = TableField(table, data_type=FieldType.String, field_id=mock_table_field_data['id'], data=mock_table_field_data)
    domain = {
        'min': 10,
        'max': 20,
        'items': {}
    }
    field.domain = domain
    assert field.domain == domain


def test_field_delete(api, mock_table_data, mock_table_field_data):
    table = Table(api, mock_table_data['uuid'], mock_table_data)
    field = TableField(table, data_type=FieldType.String, field_id=mock_table_field_data['id'], data=mock_table_field_data)
    endpoint = field.endpoint
    field.delete()
    api.delete.assert_called_once_with(endpoint)


def test_field_update(api, mock_table_data, mock_table_field_data):
    table = Table(api, mock_table_data['uuid'], mock_table_data)
    field = TableField(table, data_type=FieldType.String, field_id=mock_table_field_data['id'], data=mock_table_field_data)
    updated_data = {
        'name': 'updated_name',
        'display_name': 'updated display name',
        'description': 'updated description',
        'domain': {'min': 11, 'max': 22},
        'hyperlink': True
    }
    api.put.return_value = updated_data
    result = field.update(**updated_data)
    for key, value in updated_data.items():
        assert field.data[key] == value

    assert result == updated_data
    api.put.assert_called_once_with(field.endpoint, updated_data)


def test_field_update_domain(api, mock_table_data, mock_table_field_data):
    table = Table(api, mock_table_data['uuid'], mock_table_data)
    field = TableField(table, data_type=FieldType.String, field_id=mock_table_field_data['id'], data=mock_table_field_data)
    field.domain = None
    domain = {'min': 15, 'max': 25, 'items': {'1': 'test1', '2': 'test2'}}
    api.put.return_value = {
        **mock_table_field_data,
        **domain
    }
    result = field.update_domain(range_domain={'min': 15, 'max': 25}, list_domain={'1': 'test1', '2': 'test2'})
    assert result == field.domain


def test_field_to_async(api, async_api, mock_table_data, mock_table_field_data):
    table = Table(api, uuid=mock_table_data['uuid'], data=mock_table_data)
    field = TableField(table, data_type=FieldType.String, field_id=mock_table_field_data['id'], data=mock_table_field_data)
    async_instance = field.to_async(async_api)
    assert async_instance.api == async_api
    assert type(async_instance) == AsyncTableField


def test_share(api, mock_table_data, mock_user_data):
    """Test file sharing."""
    with patch.object(api, 'post'):
        table = Table(api, uuid=mock_table_data['uuid'], data=mock_table_data)
        users = [
            User(api, user_id=mock_user_data['1']['id'], data=mock_user_data['1']),
            User(api, user_id=mock_user_data['2']['id'], data=mock_user_data['2'])
        ]
        table.share(users=users)
        api.post.assert_called_once_with(
            f'{table.endpoint}share/',
            {'user_ids': [1, 2]}, is_json=False
        )


def test_unshare(api, mock_table_data, mock_user_data):
    """Test file unsharing."""
    with patch.object(api, 'post'):
        table = Table(api, uuid=mock_table_data['uuid'], data=mock_table_data)
        users = [
            User(api, user_id=mock_user_data['1']['id'], data=mock_user_data['1']),
            User(api, user_id=mock_user_data['2']['id'], data=mock_user_data['2'])
        ]
        table.unshare(users=users)
        
        # Verify the API call
        api.post.assert_called_once_with(
            f'{table.endpoint}unshare/',
            {'user_ids': [1, 2]}, is_json=False
        )


def test_get_shared_users(api, mock_table_data, mock_user_data):
    """Test getting shared users."""
    mock_response = [
        mock_user_data['1'],
        mock_user_data['2']
    ]
    
    with patch.object(api, 'get', return_value=mock_response):
        table = Table(api, uuid=mock_table_data['uuid'], data=mock_table_data)
        result = table.get_shared_users(search='user', limit=2)

        api.get.assert_called_once_with(
            f'{table.endpoint}shared-with-users/?search=user&skip=0&limit=2'
        )

        assert len(result) == 2
        assert result[0].first_name == 'test 1'
        assert result[0].last_name == 'test 1'
        assert result[1].first_name == 'test 2'
        assert result[1].last_name == 'test 2'


def test_to_async(api, async_api, mock_table_data):
    table = Table(api, uuid=mock_table_data['uuid'], data=mock_table_data)
    async_instance = table.to_async(async_api)
    assert async_instance.api == async_api


import pytest
from unittest.mock import patch

from geobox.settings import SystemSettings
from geobox.user import User
from geobox.enums import MaxLogPolicy, InvalidDataPolicy, LoginFailurePolicy, MaxConcurrentSessionPolicy


def test_init(api, mock_system_settings_data):
    settings = SystemSettings(api, mock_system_settings_data)
    assert settings.data == mock_system_settings_data


def test_properties(api, mock_system_settings_data):
    settings = SystemSettings(api, mock_system_settings_data)
    assert settings.max_log_policy == MaxLogPolicy(mock_system_settings_data['max_log_policy'])
    assert settings.invalid_data_policy == InvalidDataPolicy(mock_system_settings_data['invalid_data_policy'])
    assert settings.login_failure_policy == LoginFailurePolicy(mock_system_settings_data['login_failure_policy'])
    assert settings.max_concurrent_session_policy == MaxConcurrentSessionPolicy(mock_system_settings_data['max_concurrent_session_policy'])


def test_repr(api, mock_system_settings_data):
    settings = SystemSettings(api, mock_system_settings_data)
    assert repr(settings) == 'SystemSettings()'


def test_get_system_settings(api, mock_system_settings_data):
    api.get.return_value = mock_system_settings_data
    settings = SystemSettings.get_system_settings(api)
    api.get.assert_called_once_with(f'{SystemSettings.BASE_ENDPOINT}?f=json')
    assert type(settings) == SystemSettings
    assert settings.data == mock_system_settings_data


def test_update(api, mock_system_settings_data):
    settings = SystemSettings(api, mock_system_settings_data)
    updated_data = {**mock_system_settings_data,
                    'max_log': 100,
                    'login_failure_policy': LoginFailurePolicy.DisableAccount.value}

    api.put.return_value = updated_data
    settings.update(max_log=100, login_failure_policy=LoginFailurePolicy.DisableAccount)
    api.put.assert_called_once_with(settings.BASE_ENDPOINT, {'max_log': 100, 'login_failure_policy': 'DisableAccount'})


def test_to_async(api, async_api, mock_system_settings_data):
    settings = SystemSettings(api, mock_system_settings_data)
    async_instance = settings.to_async(async_api)
    assert async_instance.api == async_api


import pytest
from unittest import mock
from unittest.mock import patch, PropertyMock
from geobox.file import File
from geobox.task import Task
import itertools
from datetime import datetime

from geobox.enums import LayerType, TaskStatus
from geobox.vectorlayer import VectorLayer
from geobox.raster import Raster
from geobox.model3d import Model
from geobox.tile3d import Tile3d


def test_init(api, mock_success_task_data):
    """Test Task initialization."""
    task = Task(api, uuid=mock_success_task_data['uuid'], data=mock_success_task_data)
    assert task.uuid == mock_success_task_data['uuid']
    assert task.data == mock_success_task_data
    assert task.endpoint == f"{Task.BASE_ENDPOINT}{mock_success_task_data['uuid']}/"


def test_repr(api, mock_success_task_data):
    """Test Task string representation."""
    task = Task(api, uuid=mock_success_task_data['uuid'], data=mock_success_task_data)
    assert repr(task) == f"Task(uuid={task.uuid}, name={mock_success_task_data['name']})"


def test_output_asset_property(api, mock_publish_task_data, mock_vector_data, mock_raster_data, mock_model_data, mock_file_data, mock_tile3d_data):
    # raster layer
    task = Task(api, uuid=mock_publish_task_data['uuid'], data=mock_publish_task_data)
    api.get_raster.return_value = Raster(api, mock_raster_data['uuid'], data=mock_raster_data)
    expected_layer = Raster(api, mock_raster_data['uuid'], data=mock_raster_data)
    assert task.output_asset.name == expected_layer.name
    assert task.output_asset.uuid == expected_layer.uuid
    assert task.output_asset.data == expected_layer.data
    assert type(task.output_asset) == Raster
    # vector layer
    mock_publish_task_data['result']['layer_uuid'] = mock_publish_task_data['result']['raster_uuid']
    del mock_publish_task_data['result']['raster_uuid']
    task = Task(api, uuid=mock_publish_task_data['uuid'], data=mock_publish_task_data)
    api.get_vector.return_value = VectorLayer(api, mock_vector_data['uuid'], layer_type=LayerType.Polygon, data=mock_vector_data)
    expected_layer = VectorLayer(api, mock_vector_data['uuid'], layer_type=LayerType.Polygon, data=mock_vector_data)
    assert task.output_asset.name == expected_layer.name
    assert task.output_asset.layer_type == expected_layer.layer_type
    assert task.output_asset.uuid == expected_layer.uuid
    assert task.output_asset.data == expected_layer.data
    assert type(task.output_asset) == VectorLayer
    #model
    mock_publish_task_data['result']['model_uuid'] = mock_publish_task_data['result']['layer_uuid']
    del mock_publish_task_data['result']['layer_uuid']
    task = Task(api, uuid=mock_publish_task_data['uuid'], data=mock_publish_task_data)
    api.get_model.return_value = Model(api, mock_model_data['uuid'], data=mock_model_data)
    expected_layer = Model(api, mock_model_data['uuid'], data=mock_model_data)
    assert task.output_asset.name == expected_layer.name
    assert task.output_asset.uuid == expected_layer.uuid
    assert task.output_asset.data == expected_layer.data
    assert type(task.output_asset) == Model
    # file
    mock_publish_task_data['result']['file_uuid'] = mock_publish_task_data['result']['model_uuid']
    del mock_publish_task_data['result']['model_uuid']
    task = Task(api, uuid=mock_publish_task_data['uuid'], data=mock_publish_task_data)
    api.get_file.return_value = File(api, mock_file_data['uuid'], data=mock_file_data)
    expected_layer = File(api, mock_file_data['uuid'], data=mock_file_data)
    assert task.output_asset.name == expected_layer.name
    assert task.output_asset.uuid == expected_layer.uuid
    assert task.output_asset.data == expected_layer.data
    assert type(task.output_asset) == File
    # tile3d
    mock_publish_task_data['result']['3dtiles_uuid'] = mock_publish_task_data['result']['file_uuid']
    del mock_publish_task_data['result']['file_uuid']
    task = Task(api, uuid=mock_publish_task_data['uuid'], data=mock_publish_task_data)
    api.get_3dtile.return_value = Tile3d(api, mock_tile3d_data['uuid'], data=mock_tile3d_data)
    expected_layer = Tile3d(api, mock_tile3d_data['uuid'], data=mock_tile3d_data)
    assert task.output_asset.name == expected_layer.name
    assert task.output_asset.uuid == expected_layer.uuid
    assert task.output_asset.data == expected_layer.data
    assert type(task.output_asset) == Tile3d
    #None
    mock_publish_task_data['name'] = 'import'
    del mock_publish_task_data['result']['3dtiles_uuid']
    task = Task(api, uuid=mock_publish_task_data['uuid'], data=mock_publish_task_data)
    assert task.output_asset == None


def test_properties_success(api, mock_success_task_data):
    """Test Task properties for successful task."""
    task = Task(api, uuid=mock_success_task_data['uuid'], data=mock_success_task_data)
    api.get.return_value = mock_success_task_data
    assert task.name == mock_success_task_data['name']
    assert task.status == TaskStatus.SUCCESS
    assert task.errors is None
    assert task.accepted_at == datetime.strptime(mock_success_task_data['accepted_at'], "%Y-%m-%dT%H:%M:%S.%f")
    assert task.started_at == datetime.strptime(mock_success_task_data['started_at'], "%Y-%m-%dT%H:%M:%S.%f")
    assert task.finished_at == datetime.strptime(mock_success_task_data['finished_at'], "%Y-%m-%dT%H:%M:%S.%f")
    assert task.result == mock_success_task_data['result']


def test_properties_failure(api, mock_failure_task_data):
    """Test Task properties for failed task."""
    task = Task(api, uuid=mock_failure_task_data['uuid'], data=mock_failure_task_data)
    api.get.return_value = mock_failure_task_data
    assert task.name == mock_failure_task_data['name']
    assert task.status == TaskStatus.FAILURE
    assert task.errors == {
            'detail': {
                'msg': 'raster dataset with the same name already exists.',
                'type': 'geobox.DuplicatRasterDatasetNameError'
            }
        }
    assert task.accepted_at == datetime.strptime(mock_failure_task_data['accepted_at'], "%Y-%m-%dT%H:%M:%S.%f")
    assert task.started_at == datetime.strptime(mock_failure_task_data['started_at'], "%Y-%m-%dT%H:%M:%S.%f")
    assert task.finished_at == datetime.strptime(mock_failure_task_data['finished_at'], "%Y-%m-%dT%H:%M:%S.%f")
    assert task.result == mock_failure_task_data['result']


def test_progress(api, mock_success_task_data):
    """Test Task progress property."""
    mock_response = {
        'current': 50,
        'total': 100
    }
    with patch.object(api, 'get', return_value=mock_response):
        task = Task(api, uuid=mock_success_task_data['uuid'], data=mock_success_task_data)
        assert task.progress == 50


def test_progress_not_running(api, mock_success_task_data):
    """Test Task progress property when task is not running."""
    mock_response = {
        'current': None,
        'total': None
    }
    with patch.object(api, 'get', return_value=mock_response):
        task = Task(api, uuid=mock_success_task_data['uuid'], data=mock_success_task_data)
        assert task.progress is None


def test_wait_success(api, mock_success_task_data):
    """Test Task wait method with successful completion."""
    task = Task(api, uuid=mock_success_task_data['uuid'], data=mock_success_task_data)
    states = [TaskStatus.PENDING, TaskStatus.SUCCESS]
    progress_values = [0, 50, 100]

    with patch.object(type(task), 'status', new_callable=PropertyMock) as mock_status, \
         patch.object(type(task), 'progress', new_callable=PropertyMock) as mock_progress:
        mock_status.side_effect = states
        mock_progress.side_effect = progress_values
        result = task.wait(timeout=5, interval=0.1)
        assert result == TaskStatus.SUCCESS

    import builtins
    original_import = builtins.__import__

    def mocked_import(name, globals=None, locals=None, fromlist=(), level=0):
        if name == "tqdm.auto":
            raise ImportError("No module named 'tqdm'")
        return original_import(name, globals, locals, fromlist, level)

    with mock.patch("builtins.__import__", side_effect=mocked_import):
        with mock.patch("geobox.api.logger") as mock_logger:
            result = task._create_progress_bar()
            assert result is None
            mock_logger.warning.assert_called_once()


def test_wait_failure(api, mock_failure_task_data):
    """Test Task wait method with failure."""
    task = Task(api, uuid=mock_failure_task_data['uuid'], data=mock_failure_task_data)
    states = [TaskStatus.PENDING, TaskStatus.FAILURE]
    progress_values = [0, 30, 60]
    
    with patch.object(type(task), 'status', new_callable=PropertyMock) as mock_status, \
         patch.object(type(task), 'progress', new_callable=PropertyMock) as mock_progress:
        mock_status.side_effect = states
        mock_progress.side_effect = progress_values
        result = task.wait(timeout=5, interval=0.1)
        assert result == TaskStatus.FAILURE


def test_wait_timeout_pending(api, mock_success_task_data):
    """Test Task wait method with timeout."""
    task = Task(api, uuid=mock_success_task_data['uuid'], data=mock_success_task_data)
    with patch.object(type(task), 'status', new_callable=PropertyMock) as mock_status:
        with pytest.raises(TimeoutError, match=f"Task {mock_success_task_data['name']} timed out after 1 seconds"):
            mock_status.return_value = 'PENDING'
            result = task.wait(timeout=1, interval=0.1)
            assert result == task


def test_wait_timeout(api, mock_success_task_data):
    """Test Task wait method raises TimeoutError when task times out."""
    task = Task(api, uuid=mock_success_task_data['uuid'], data=mock_success_task_data)
    with patch.object(type(task), 'status', new_callable=PropertyMock) as mock_status, \
         patch.object(type(task), 'progress', new_callable=PropertyMock) as mock_progress:
        mock_status.return_value = TaskStatus.PENDING
        mock_progress.side_effect = itertools.cycle([0, 10, 20, 30, 40, 50])  
        with pytest.raises(TimeoutError, match=f"Task {mock_success_task_data['name']} timed out after 1 seconds"):
            task.wait(timeout=1, interval=0.1)


def test_wait_without_progress_bar(api, mock_success_task_data):
    """Test Task wait method without progress bar."""
    task = Task(api, uuid=mock_success_task_data['uuid'], data=mock_success_task_data)
    states = [TaskStatus.PENDING, TaskStatus.SUCCESS]
    progress_values = [0, 50, 100]
    
    with patch.object(type(task), 'status', new_callable=PropertyMock) as mock_status, \
         patch.object(type(task), 'progress', new_callable=PropertyMock) as mock_progress, \
         patch('tqdm.auto.tqdm') as mock_tqdm:
        mock_status.side_effect = states
        mock_progress.side_effect = progress_values
        result = task.wait(timeout=5, interval=0.1, progress_bar=False)
        assert result == TaskStatus.SUCCESS
        mock_tqdm.assert_not_called()


def test_abort(api, mock_success_task_data):
    """Test Task abort method."""
    with patch.object(api, 'post') as mock_post:
        task = Task(api, uuid=mock_success_task_data['uuid'], data=mock_success_task_data)
        task.abort()
        mock_post.assert_called_once_with(f"tasks/{mock_success_task_data['uuid']}/abort/")


def test_get_tasks(api, mock_success_task_data):
    """Test getting list of tasks."""
    mock_response = [
        mock_success_task_data,
        mock_success_task_data
    ]

    with patch.object(api, 'get', return_value=mock_response):
        tasks = Task.get_tasks(api, search_fields='name', search='test')
        assert len(tasks) == 2
        assert all(isinstance(t, Task) for t in tasks)
        assert tasks[0].uuid == mock_success_task_data['uuid']
        assert tasks[1].uuid == mock_success_task_data['uuid']
    # return count
    mock_response = 42
    with patch.object(api, 'get', return_value=mock_response):
        count = Task.get_tasks(api, return_count=True)
        assert count == 42


def test_get_task(api, mock_success_task_data):
    """Test getting a single task."""
    with patch.object(api, 'get', return_value=mock_success_task_data):
        task = Task.get_task(api, uuid=mock_success_task_data['uuid'])
        assert isinstance(task, Task)
        assert task.uuid == mock_success_task_data['uuid']
        assert task.name == mock_success_task_data['name']
        assert task.data == mock_success_task_data


def test_to_async(api, async_api, mock_success_task_data):
    task = Task(api, uuid=mock_success_task_data['uuid'], data=mock_success_task_data)
    async_instance = task.to_async(async_api)
    assert async_instance.api == async_api


import pytest
from unittest.mock import Mock, patch

from geobox.query import Query, QueryGeometryType, QueryParamType, QueryResultType
from geobox.task import Task
from geobox.user import User


def test_init(api, mock_query_data):
    """Test Query initialization."""
    query = Query(api, mock_query_data['uuid'], mock_query_data)
    
    assert query.api == api
    assert query.uuid == mock_query_data['uuid']
    assert query.data == mock_query_data
    assert query.result == {}
    assert query._system_query == False


def test_check_access_system_query(api, mock_query_data):
    """Test _check_access raises exception for system queries."""
    query = Query(api, mock_query_data['uuid'], mock_query_data)
    query._system_query = True
    
    with pytest.raises(PermissionError, match="Cannot modify system queries - they are read-only"):
        query._check_access()


def test_check_access_regular_query(api, mock_query_data):
    """Test _check_access allows regular queries."""
    query = Query(api, mock_query_data['uuid'], mock_query_data)
    query._system_query = False
    
    query._check_access()


def test_get_sql(api, mock_query_data):
    """Test sql property getter."""
    query = Query(api, mock_query_data['uuid'], mock_query_data)
    assert query.sql == mock_query_data['sql']


def test_set_sql(api, mock_query_data):
    """Test sql property setter."""
    query = Query(api, mock_query_data['uuid'], mock_query_data)
    new_sql = "SELECT * FROM new_table"
    
    query.sql = new_sql
    
    assert query.data['sql'] == new_sql


def test_get_params(api, mock_query_data):
    """Test params property getter."""
    query = Query(api, mock_query_data['uuid'], mock_query_data)
    
    assert query.params == mock_query_data['params']


def test_get_params_empty_list(api, mock_query_data):
    """Test params property getter initializes empty list."""
    mock_query_data = {**mock_query_data, **{'params': 100}}
    query = Query(api, mock_query_data['uuid'], mock_query_data)
    
    assert query.params == []
    assert query.data['params'] == []


def test_set_params(api, mock_query_data):
    """Test params property setter."""
    query = Query(api, mock_query_data['uuid'], mock_query_data)
    new_param = [
        {"name": "foo", "value": "bar", "type": "Layer"},
        {"name": "baz", "value": "qux", "type": "Layer"}
    ]
    query.params = new_param
    assert query.params == new_param


def test_set_params_triggers_initialization(api, mock_query_data):
    """Trigger line 131 in setter by setting params when initial params is not a list."""
    # Set initial params to a non-list value
    mock_query_data = {**mock_query_data, 'params': 100}  # integer, not list
    query = Query(api, mock_query_data['uuid'], mock_query_data)
    
    # Now assign to params with any value, line 131 should execute
    query.params = [{"name": "foo", "value": "bar", "type": "Layer"}]
    
    assert isinstance(query.data['params'], list)


@patch('geobox.map.urljoin')
@patch('geobox.map.urlencode')
def test_get_queries(mock_urlencode, mock_urljoin, mock_query_data, api):
    """Test get_queries class method."""
    mock_urlencode.return_value = 'f=json'
    mock_urljoin.return_value = 'queries/?f=json'
    
    api.get.return_value = [mock_query_data]

    queries = Query.get_queries(api, uuid=mock_query_data['uuid'], data=mock_query_data)
    assert len(queries) == 1
    assert queries[0].name == mock_query_data['name']
    assert queries[0].uuid == mock_query_data['uuid']
    assert queries[0].data == mock_query_data
    api.get.assert_called_once_with(f'{Query.BASE_ENDPOINT}?f=json&return_count=False&skip=0&limit=10&shared=False')

    # Test with return_count
    api.reset_mock()
    api.get.return_value = 5
    count = Query.get_queries(api, return_count=True)
    assert count == 5
    api.get.assert_called_once_with(f'{Query.BASE_ENDPOINT}?f=json&return_count=True&skip=0&limit=10&shared=False')


def test_create_query(mock_query_data, api):
    """Test create_query class method."""
    api.post.return_value = mock_query_data
    query = Query.create_query(api, 
                                name='new', 
                                sql='SELECT * FROM test', 
                                params=[{'name': 'test', 'type': 'Layer', 'value': 'test_value'}])
    assert isinstance(query, Query)
    assert query.data == mock_query_data
    api.post.assert_called_once_with(Query.BASE_ENDPOINT, {'name': 'new', 'sql': 'SELECT * FROM test', 'params': [{'name': 'test', 'type': 'Layer', 'value': 'test_value'}]})


@patch('geobox.map.urljoin')
@patch('geobox.map.urlencode')
def test_get_query(mock_urlencode, mock_urljoin, mock_query_data, api):
    """Test get_query class method."""
    mock_urlencode.return_value = 'f=json'
    mock_urljoin.return_value = f'{Query.BASE_ENDPOINT}{mock_query_data["uuid"]}/?f=json'
    api.get.return_value = mock_query_data

    query = Query.get_query(api, mock_query_data['uuid'])
    api.get.assert_called_once_with(f'{query.endpoint}?f=json')
    assert query.data == mock_query_data


@patch('geobox.map.urljoin')
@patch('geobox.map.urlencode')
def test_get_query_by_name(mock_urlencode, mock_urljoin, mock_query_data, api):
    """Test get_query class method."""
    mock_urlencode.return_value = 'f=json'
    mock_urljoin.return_value = f'{Query.BASE_ENDPOINT}{mock_query_data["uuid"]}/?f=json'
    api.get.return_value = [mock_query_data]

    query = Query.get_query_by_name(api, name=mock_query_data['name'])
    api.get.assert_called_once_with(f'{Query.BASE_ENDPOINT}?f=json&q=name+%3D+%27new%27&return_count=False&skip=0&limit=10&shared=False')
    assert query.data == mock_query_data
    # not found
    query = Query.get_query_by_name(api, name='not_found')
    assert query == None


def test_get_system_queries(mock_query_data, api):
    """Test get_system_queries class method."""
    api.get.return_value = [mock_query_data, mock_query_data, mock_query_data]
    
    result = Query.get_system_queries(api)
    
    api.get.assert_called_once_with(f'{Query.BASE_ENDPOINT}systemQueries/?f=json&return_count=False&skip=0&limit=100&shared=False')
    assert len(result) == 3
    assert isinstance(result[0], Query)
    assert result[0]._system_query == True


def test_add_param(api, mock_query_data):
    """Test add_param method."""
    query = Query(api, mock_query_data['uuid'], mock_query_data)
    original_length = len(query.params)
    
    query.add_param(
        name="test_param",
        value="test_value",
        type=QueryParamType.TEXT,
        default_value="default",
        Domain={'key': 'value'}
    )
    
    assert len(query.params) == original_length + 1
    new_param = query.params[-1]
    assert new_param['name'] == "test_param"
    assert new_param['value'] == "test_value"
    assert new_param['type'] == QueryParamType.TEXT.value
    assert new_param['default_value'] == "default"
    assert new_param['Domain'] == {'key': 'value'}



def test_remove_param(api, mock_query_data):
    """Test remove_param method."""
    query = Query(api, mock_query_data['uuid'], mock_query_data)
    original_length = len(query.params)
    param_name = query.params[0]['name']
    
    query.remove_param(param_name)
    
    assert len(query.params) == original_length - 1

    # not found
    query = Query(api, mock_query_data['uuid'], mock_query_data)
    with pytest.raises(ValueError, match="Parameter with name 'nonexistent' not found"):
        query.remove_param("nonexistent")



def test_execute(api, mock_query_data):
    """Test execute method."""
    api.post.return_value = {'result': 'success'}
    query = Query(api, mock_query_data['uuid'], mock_query_data)
    
    result = query.execute(
        f="json",
        result_type=QueryResultType.both,
        return_count=True,
        out_srid=4326,
        skip=10,
        limit=100
    )
    
    api.post.assert_called_once_with(f'{query.BASE_ENDPOINT}exec/', {'f': 'json', 'sql': 'SELECT layer.*, ST_Centroid(geom) as goem\nFROM layer', 'params': [{'name': 'layer', 'type': 'Layer', 'default_value': None, 'domain': None}, {'name': 'layer', 'value': '297fa7ca-877a-400c-8003-d65de9e791c2', 'type': 'Layer', 'default_value': '297fa7ca-877a-400c-8003-d65de9e791c2', 'Domain': {}}], 'result_type': 'both', 'return_count': True, 'out_srid': 4326, 'quant_factor': 1000000, 'skip': 10, 'limit': 100, 'skip_geometry': False})
    assert result == {'result': 'success'}
    assert query.result == {'result': 'success'}


def test_execute_no_sql(api, mock_query_data):
    """Test execute raises exception when no SQL."""
    data = mock_query_data.copy()
    data['sql'] = None
    query = Query(api, data['uuid'], data)
    
    with pytest.raises(ValueError, match='"sql" parameter is required'):
        query.execute()


def test_execute_no_params(api, mock_query_data):
    """Test execute raises exception when no params."""
    mock_query_data['params'] = []
    query = Query(api, mock_query_data['uuid'], mock_query_data)
    
    with pytest.raises(ValueError, match='"params" parameter is required'):
        query.execute()


def test_update(api, mock_query_data):
    """Test update_query method."""
    query = Query(api, mock_query_data['uuid'], mock_query_data)
    updated_query_data = mock_query_data.copy()
    updated_query_data['name'] = "new_name"
    updated_query_data['display_name'] = "New Display Name"
    updated_query_data['sql'] = "SELECT * FROM new_table"
    updated_query_data['params'] = [{'name': 'new_param'}]

    api.put.return_value = updated_query_data
    query.update(
        name="new_name",
        display_name="New Display Name",
        sql="SELECT * FROM new_table",
        params=[{'name': 'new_param'}]
    )
    
    api.put.assert_called_once_with(
        query.endpoint,
        {
            "name": "new_name",
            "display_name": "New Display Name",
            "sql": "SELECT * FROM new_table",
            "params": [{'name': 'new_param'}]
        }
    )
    assert query.data == updated_query_data

    # update system queries
    query = Query(api, mock_query_data['uuid'], mock_query_data)
    query._system_query = True
    
    with pytest.raises(PermissionError):
        query.update(name="test")


def test_save_existing_query(api, mock_query_data):
    """Test save method for existing query."""
    query = Query(api, mock_query_data['uuid'], mock_query_data)
    query.save()

    api.put.assert_called_once_with(query.endpoint, {'name': 'new', 'sql': 'SELECT layer.*, ST_Centroid(geom) as goem\nFROM layer', 'params': [{'name': 'layer', 'value': '297fa7ca-877a-400c-8003-d65de9e791c2', 'type': 'Layer', 'default_value': '297fa7ca-877a-400c-8003-d65de9e791c2', 'Domain': {}}]}
)


def test_save_new_query(api, mock_query_data):
    """Test save method for new query."""
    response = {**mock_query_data, **{'uuid': mock_query_data['uuid']}}
    api.post.return_value = response
    del mock_query_data['uuid']
    query = Query(api, data=mock_query_data)
    
    query.save()
        
    api.post.assert_called_once_with(query.BASE_ENDPOINT, mock_query_data)
    assert query.data == response


def test_save_removes_empty_params(api, mock_query_data):
    """Test save method removes empty params."""
    query_data_copy = mock_query_data.copy()
    query_data_copy['params'] = [
        {'name': 'param1', 'value': 'value1'},
        {'name': 'param2', 'value': ''},
        {'name': 'param3', 'value': None}
    ]
    query = Query(api, mock_query_data['uuid'], query_data_copy)
    
    query.save()
    
    # Should remove params without value
    assert len(query.params) == 1
    assert query.params[0]['name'] == 'param1'


def test_delete(api, mock_query_data):
    """Test delete method."""
    query = Query(api, mock_query_data['uuid'], mock_query_data)
    endpoint = query.endpoint
    query.delete()
    api.delete.assert_called_once_with(endpoint)

    # delete system query
    query = Query(api, mock_query_data['uuid'], mock_query_data)
    query._system_query = True
    
    with pytest.raises(PermissionError):
        query.delete()


def test_share(api, mock_query_data, mock_user_data):
    """Test file sharing."""
    with patch.object(api, 'post'):
        query = Query(api, mock_query_data['uuid'], mock_query_data)
        users = [
            User(api, user_id=mock_user_data['1']['id'], data=mock_user_data['1']),
            User(api, user_id=mock_user_data['2']['id'], data=mock_user_data['2'])
        ]
        query.share(users=users)
        api.post.assert_called_once_with(
            f'{query.endpoint}share/',
            {'user_ids': [1, 2]}, is_json=False
        )
        # share system queries
        query = Query(api, mock_query_data['uuid'], mock_query_data)
        query._system_query = True
    
        with pytest.raises(PermissionError):
            query.share([Mock(spec=User)])


def test_unshare(api, mock_query_data, mock_user_data):
    """Test file unsharing."""
    with patch.object(api, 'post'):
        query = Query(api, mock_query_data['uuid'], mock_query_data)
        users = [
            User(api, user_id=mock_user_data['1']['id'], data=mock_user_data['1']),
            User(api, user_id=mock_user_data['2']['id'], data=mock_user_data['2'])
        ]
        query.unshare(users=users)
        api.post.assert_called_once_with(
            f'{query.endpoint}unshare/',
            {'user_ids': [1, 2]}, is_json=False
        )
        # unshare system queries
        query = Query(api, mock_query_data['uuid'], mock_query_data)
        query._system_query = True
    
        with pytest.raises(PermissionError):
            query.unshare([Mock(spec=User)])


def test_get_shared_users(api, mock_query_data, mock_user_data):
    """Test getting shared users."""
    mock_response = [
        mock_user_data['1'],
        mock_user_data['2']
    ]
    
    with patch.object(api, 'get', return_value=mock_response):
        query = Query(api, uuid=mock_query_data['uuid'], data=mock_query_data)
        result = query.get_shared_users(search='user', limit=2)

        api.get.assert_called_once_with(
            f'{query.endpoint}shared-with-users/?search=user&skip=0&limit=2'
        )

        assert len(result) == 2
        assert result[0].first_name == 'test 1'
        assert result[0].last_name == 'test 1'
        assert result[1].first_name == 'test 2'
        assert result[1].last_name == 'test 2'

        # get system query shared users
        query = Query(api, mock_query_data['uuid'], mock_query_data)
        query._system_query = True
    
        with pytest.raises(PermissionError):
            query.get_shared_users(search='user', limit=2)


def test_thumbnail(api, mock_query_data):
    """Test thumbnail property."""
    query = Query(api, mock_query_data['uuid'], mock_query_data)
    
    thumbnail_url = query.thumbnail
    
    assert thumbnail_url == f'{api.base_url}{query.endpoint}thumbnail.png'


def test_save_as_layer(api, mock_query_data, mock_success_task_data):
    """Test save_as_layer method."""
    api.post.return_value = {'task_id': mock_success_task_data['id']}
    api.get.return_value = mock_success_task_data
    query = Query(api, mock_query_data['uuid'], mock_query_data)
    
    task = query.save_as_layer("new_layer", QueryGeometryType.POINT)
    
    api.post.assert_called_once_with(f'{query.BASE_ENDPOINT}saveAsLayer/', {'sql': 'SELECT layer.*, ST_Centroid(geom) as goem\nFROM layer', 'params': [{'name': 'layer', 'type': 'Layer', 'value': None}, {'name': 'layer', 'type': 'Layer', 'value': '297fa7ca-877a-400c-8003-d65de9e791c2'}], 'layer_name': 'new_layer', 'layer_type': 'Point'})
    assert isinstance(task, Task)


def test_to_async(api, async_api, mock_query_data):
    query = Query(api, mock_query_data['uuid'], mock_query_data)
    async_instance = query.to_async(async_api)
    assert async_instance.api == async_api 


import pytest
from unittest.mock import patch

from geobox.apikey import ApiKey


def test_init(api, mock_apikey_data):
    apikey = ApiKey(api, mock_apikey_data['id'], mock_apikey_data)
    assert apikey.name == mock_apikey_data['name']
    assert apikey.key_id == mock_apikey_data['id']
    assert apikey.data == mock_apikey_data
    assert apikey.endpoint == f'{ApiKey.BASE_ENDPOINT}{apikey.key_id}'


def test_repr(api, mock_apikey_data):
    apikey = ApiKey(api, mock_apikey_data['id'], mock_apikey_data)
    assert repr(apikey) == f'ApiKey(id={apikey.id}, name={apikey.name}, revoked={apikey.revoked})'


def test_get_apikeys(api, mock_apikey_data):
    api.get.return_value = [mock_apikey_data, mock_apikey_data]
    apikeys = ApiKey.get_apikeys(api)
    api.get.assert_called_once_with(f'{ApiKey.BASE_ENDPOINT}')
    assert len(apikeys) == 2
    assert type(apikeys[0]) == ApiKey
    assert apikeys[0].data == mock_apikey_data


def test_create_apikey(api, mock_apikey_data):
    api.post.return_value = mock_apikey_data
    apikey = ApiKey.create_apikey(api, name=mock_apikey_data['name'])
    api.post.assert_called_once_with(ApiKey.BASE_ENDPOINT, payload={'name': mock_apikey_data['name']}, is_json=False)
    assert type(apikey) == ApiKey
    assert apikey.key_id == mock_apikey_data['id']
    assert apikey.data == mock_apikey_data


def test_get_apikey(api, mock_apikey_data):
    api.get.return_value = mock_apikey_data
    apikey = ApiKey.get_apikey(api, key_id=mock_apikey_data['id'])
    api.get.assert_called_once_with(f"{ApiKey.BASE_ENDPOINT}{mock_apikey_data['id']}/?f=json")
    assert type(apikey) == ApiKey
    assert apikey.key_id == mock_apikey_data['id']
    assert apikey.data == mock_apikey_data


def test_get_apikey_by_name(api, mock_apikey_data):
    api.get.return_value = [mock_apikey_data]
    apikey = ApiKey.get_apikey_by_name(api, name=mock_apikey_data['name'])
    api.get.assert_called_once_with(f"{ApiKey.BASE_ENDPOINT}?search=test")
    assert type(apikey) == ApiKey
    assert apikey.key_id == mock_apikey_data['id']
    assert apikey.data == mock_apikey_data
    # not found
    apikey = ApiKey.get_apikey_by_name(api, name='invalid_name')
    assert apikey == None


def test_update(api, mock_apikey_data):
    apikey = ApiKey(api, mock_apikey_data['id'], mock_apikey_data)
    updated_data = {**mock_apikey_data,
                    'name': 'updated_name'}

    api.put.return_value = updated_data
    apikey.update(name='updated_name')
    api.put.assert_called_once_with(apikey.endpoint, {'name': 'updated_name'}, is_json=False)


def test_delete(api, mock_apikey_data):
    apikey = ApiKey(api, mock_apikey_data['id'], mock_apikey_data)
    endpoint = apikey.endpoint
    apikey.delete()
    api.delete.assert_called_once_with(endpoint)
    assert apikey.key_id is None
    assert apikey.endpoint is None


def test_revoke(api, mock_apikey_data):
    apikey = ApiKey(api, mock_apikey_data['id'], mock_apikey_data)
    apikey.revoke()
    api.post.assert_called_once_with(f'{apikey.endpoint}/revoke')
    assert apikey.revoked == True


def test_grant(api, mock_apikey_data):
    apikey = ApiKey(api, mock_apikey_data['id'], mock_apikey_data)
    apikey.grant()
    api.post.assert_called_once_with(f'{apikey.endpoint}/grant')
    assert apikey.revoked == False


def test_to_async(api, async_api, mock_apikey_data):
    apikey = ApiKey(api, mock_apikey_data['id'], mock_apikey_data)
    async_instance = apikey.to_async(async_api)
    assert async_instance.api == async_api  